from argparse import Namespace
from asyncio import to_thread
from ...cli import get_input
from ...cli.commands import get_model_settings
from ...entities import DistanceType, Modality, SearchMatch, Similarity
from ...cli.commands.model import model_display
from ...memory.partitioner.text import TextPartitioner, TextPartition
from ...memory.partitioner.code import CodePartitioner
from ...memory.permanent import MemoryType
from ...memory.permanent.pgsql.raw import PgsqlRawMemory
from uuid import UUID
from ...model.hubs.huggingface import HuggingfaceHub
from ...model.manager import ModelManager
from faiss import IndexFlatL2
from httpx import AsyncClient, Response
from pathlib import Path
from urllib.parse import urlparse
from io import BytesIO
from logging import Logger
from markitdown import MarkItDown, DocumentConverterResult
from numpy import abs, corrcoef, dot, sum, vstack
from numpy.linalg import norm
from rich.console import Console
from rich.theme import Theme


async def memory_document_index(
    args: Namespace,
    console: Console,
    theme: Theme,
    hub: HuggingfaceHub,
    logger: Logger,
) -> None:
    assert args.model and args.source and args.partition_max_tokens
    assert args.partition_overlap and args.partition_window
    assert args.dsn and args.participant and args.namespace

    def transform(html: bytes) -> DocumentConverterResult:
        return MarkItDown().convert_stream(BytesIO(html))

    _, _i = theme._, theme.icons
    model_id = args.model
    source = args.source
    participant_id = UUID(args.participant)
    namespace = args.namespace
    dsn = args.dsn
    identifier = args.identifier or None
    display_partitions = (
        args.display_partitions if not args.no_display_partitions else None
    )

    with ModelManager(hub, logger) as manager:
        engine_uri = manager.parse_uri(args.model)
        model_settings = get_model_settings(
            args, hub, logger, engine_uri, modality=Modality.EMBEDDING
        )

        with manager.load(**model_settings) as stm:
            logger.debug("Loaded model %s", stm.config.__repr__())

            model_display(
                args, console, theme, hub, logger, model=stm, summary=True
            )

            contents: str | None = None

            is_url = urlparse(source).scheme in ("http", "https")
            if is_url:
                async with AsyncClient() as client:
                    response: Response = await client.get(source)
                    response.raise_for_status()
                    result = await to_thread(transform, response.content)
                    contents = result.text_content
            else:
                contents = Path(source).read_text(encoding=args.encoding)

            if not identifier:
                identifier = source if is_url else str(Path(source).resolve())

            if is_url:
                memory_type = MemoryType.URL
            else:
                memory_type = (
                    MemoryType.CODE
                    if args.partitioner == "code"
                    else MemoryType.FILE
                )

            if is_url or args.partitioner == "text":
                partitioner = TextPartitioner(
                    stm,
                    logger,
                    max_tokens=args.partition_max_tokens,
                    window_size=args.partition_window,
                    overlap_size=args.partition_overlap,
                )
                partitions = await partitioner(contents)
            else:
                code_partitioner = CodePartitioner(logger)
                code_partitions, _ = await to_thread(
                    code_partitioner.partition,
                    args.language or "python",
                    contents,
                    args.encoding,
                    args.partition_max_tokens,
                )
                partitions: list[TextPartition] = []
                for cp in code_partitions:
                    embeddings = await stm(cp.data)
                    tokens = stm.token_count(cp.data)
                    partitions.append(
                        TextPartition(
                            data=cp.data,
                            embeddings=embeddings,
                            total_tokens=tokens,
                        )
                    )

            memory_store = await PgsqlRawMemory.create_instance(
                dsn=dsn, logger=logger
            )
            await memory_store.append_with_partitions(
                namespace,
                participant_id,
                memory_type=memory_type,
                data=contents,
                identifier=identifier,
                partitions=partitions,
                symbols={},
                model_id=model_id,
            )

            if display_partitions:
                console.print(
                    theme.memory_partitions(
                        partitions,
                        display_partitions=display_partitions,
                    )
                )


async def memory_embeddings(
    args: Namespace,
    console: Console,
    theme: Theme,
    hub: HuggingfaceHub,
    logger: Logger,
) -> None:
    assert args.model
    _, _i = theme._, theme.icons
    model_id = args.model
    display_partitions = (
        args.display_partitions if not args.no_display_partitions else None
    )
    compare_strings = args.compare or None
    searches = args.search or None
    search_k = args.search_k or 1
    sort_by: DistanceType = args.sort or DistanceType.L2

    sort_key = {
        DistanceType.COSINE: lambda s: s.cosine_distance,
        DistanceType.DOT: lambda s: s.inner_product,
        DistanceType.L1: lambda s: s.l1_distance,
        DistanceType.L2: lambda s: s.l2_distance,
        DistanceType.PEARSON: lambda s: s.pearson,
    }[sort_by]
    reverse_sort = sort_by in (DistanceType.COSINE, DistanceType.PEARSON)

    model_settings = get_model_settings(
        args, hub, logger, model_id, modality=Modality.EMBEDDING
    )
    with ModelManager(hub, logger) as manager:
        with manager.load(**model_settings) as stm:
            logger.debug("Loaded model %s", stm.config.__repr__())

            model_display(
                args, console, theme, hub, logger, model=stm, summary=True
            )

            input_string = get_input(
                console,
                _i["user_input"] + " ",
                echo_stdin=not args.no_repl,
                is_quiet=args.quiet,
            )
            if not input_string:
                return

            partitioner = (
                TextPartitioner(
                    stm,
                    logger,
                    max_tokens=args.partition_max_tokens,
                    window_size=args.partition_window,
                    overlap_size=args.partition_overlap,
                )
                if args.partition
                else None
            )

            logger.debug(
                f'Looking to embed string "{input_string}" with {model_id}'
            )

            input_strings = (
                [input_string, *compare_strings]
                if compare_strings
                else input_string
            )

            embeddings = await stm(input_strings)

            input_string_embeddings = (
                embeddings[0] if compare_strings else embeddings
            )
            total_tokens = stm.token_count(input_string)

            # Subject string
            if partitioner and display_partitions:
                partitions = await partitioner(input_string)

                console.print(
                    theme.memory_partitions(
                        partitions, display_partitions=display_partitions
                    )
                )
            else:
                console.print(
                    theme.memory_embeddings(
                        input_string,
                        input_string_embeddings,
                        total_tokens=total_tokens,
                        minv=input_string_embeddings.min().item(),
                        maxv=input_string_embeddings.max().item(),
                        meanv=input_string_embeddings.mean().item(),
                        stdv=input_string_embeddings.std().item(),
                        normv=norm(input_string_embeddings).item(),
                    )
                )

            # Comparisons
            if compare_strings:
                joined = '", "'.join(compare_strings)
                logger.debug(
                    f'Calculating similarities between "{input_string}" and '
                    f'["{joined}"]'
                )
                embeddings = embeddings[1:]
                comparisons = dict(zip(compare_strings, embeddings))
                # Calculate similarities
                similarities: dict[str, Similarity] = {}
                for compare_string, compare_embeddings in comparisons.items():
                    dot_product = dot(
                        input_string_embeddings, compare_embeddings
                    )
                    cosine_distance_denom = norm(
                        input_string_embeddings
                    ) * norm(compare_embeddings)
                    cosine_distance = (
                        (dot_product / cosine_distance_denom).item()
                        if cosine_distance_denom != 0
                        else 1.0
                    )
                    inner_product = -1 * dot_product
                    l1_distance = sum(
                        abs(input_string_embeddings - compare_embeddings)
                    ).item()
                    l2_distance = norm(
                        input_string_embeddings - compare_embeddings
                    ).item()
                    pearson = corrcoef(
                        input_string_embeddings, compare_embeddings
                    )[0, 1].item()
                    similarities[compare_string] = Similarity(
                        cosine_distance=cosine_distance,
                        inner_product=inner_product,
                        l1_distance=l1_distance,
                        l2_distance=l2_distance,
                        pearson=pearson,
                    )

                similarities = dict(
                    sorted(
                        similarities.items(),
                        key=lambda item: sort_key(item[1]),
                        reverse=reverse_sort,
                    )
                )

                joined = '", "'.join(compare_strings)
                logger.debug(
                    f'Similarities between "{input_string}" and ["{joined}"]: '
                    + similarities.__repr__()
                )

                # Closest match
                most_similar = next(iter(similarities))

                console.print(
                    theme.memory_embeddings_comparison(
                        similarities, most_similar
                    )
                )

            if searches:
                knowledge_partitions = (
                    await partitioner(input_string) if partitioner else None
                )

                if knowledge_partitions and display_partitions:
                    console.print(
                        theme.memory_partitions(
                            knowledge_partitions,
                            display_partitions=display_partitions,
                        )
                    )

                index = IndexFlatL2(input_string_embeddings.shape[0])

                if partitioner:
                    knowledge_stack = vstack(
                        [kp.embeddings for kp in knowledge_partitions]
                    ).astype("float32", copy=False)
                    index.add(knowledge_stack)
                else:
                    index.add(
                        input_string_embeddings.reshape(1, -1).astype(
                            "float32", copy=False
                        )
                    )

                search_embeddings = await stm(searches)
                search_stack = vstack(search_embeddings).astype(
                    "float32", copy=False
                )
                distances, ids = index.search(search_stack, search_k)
                matches: list[tuple[int, int, float]] = [
                    (q_id, kn_id, float(dist))
                    for q_id, (dist_row, id_row) in enumerate(
                        zip(distances, ids)
                    )
                    for dist, kn_id in zip(dist_row, id_row)
                ]
                # smallest distance first
                matches.sort(key=lambda t: t[2])

                search_matches: list[SearchMatch] = []
                for q_id, kn_id, l2_distance in matches:
                    search_query = searches[q_id]
                    knowledge_chunk = (
                        knowledge_partitions[kn_id].data
                        if knowledge_partitions
                        else input_string if kn_id == 0 else None
                    )
                    if not knowledge_chunk:
                        continue
                    search_match = SearchMatch(
                        query=search_query,
                        match=knowledge_chunk,
                        l2_distance=l2_distance,
                    )
                    search_matches.append(search_match)

                console.print(theme.memory_embeddings_search(search_matches))


async def memory_search(
    args: Namespace,
    console: Console,
    theme: Theme,
    hub: HuggingfaceHub,
    logger: Logger,
) -> None:
    assert args.model and args.dsn and args.participant and args.namespace
    assert args.function

    _, _i = theme._, theme.icons
    model_id = args.model
    participant_id = UUID(args.participant)
    namespace = args.namespace
    dsn = args.dsn
    limit = args.limit

    input_string = get_input(
        console,
        _i["user_input"] + " ",
        echo_stdin=not args.no_repl,
        is_quiet=args.quiet,
    )
    if not input_string:
        return

    model_settings = get_model_settings(
        args, hub, logger, model_id, modality=Modality.EMBEDDING
    )

    with ModelManager(hub, logger) as manager:
        with manager.load(**model_settings) as stm:
            logger.debug("Loaded model %s", stm.config.__repr__())

            model_display(
                args, console, theme, hub, logger, model=stm, summary=True
            )

            partitioner = TextPartitioner(
                stm,
                logger,
                max_tokens=args.partition_max_tokens,
                window_size=args.partition_window,
                overlap_size=args.partition_overlap,
            )
            search_partitions = await partitioner(input_string)

            memory_store = await PgsqlRawMemory.create_instance(
                dsn=dsn, logger=logger
            )
            memories = await memory_store.search_memories(
                search_partitions=search_partitions,
                participant_id=participant_id,
                namespace=namespace,
                function=args.function,
                limit=limit,
            )

            console.print(
                theme.memory_search_matches(
                    participant_id, namespace, memories
                )
            )



from .. import di_get_logger, di_get_orchestrator
from ...agent.orchestrator import Orchestrator
from ...entities import (
    GenerationSettings,
    Message,
    MessageContentImage,
    MessageContentText,
    MessageRole,
)
from ...event import Event
from ...server.entities import (
    ChatCompletionChoice,
    ChatCompletionChunk,
    ChatCompletionChunkChoice,
    ChatCompletionChunkChoiceDelta,
    ChatCompletionRequest,
    ChatCompletionResponse,
    ChatMessage,
    ChatCompletionUsage,
    ContentImage,
    ContentText,
)
from fastapi import APIRouter, Depends
from fastapi.responses import StreamingResponse
from logging import Logger
from time import time
from uuid import uuid4

router = APIRouter(
    prefix="/chat",
    tags=["completions"],
)


@router.post("/completions", response_model=ChatCompletionResponse)
async def create_chat_completion(
    request: ChatCompletionRequest,
    logger: Logger = Depends(di_get_logger),
    orchestrator: Orchestrator = Depends(di_get_orchestrator),
):
    assert orchestrator and isinstance(orchestrator, Orchestrator)
    assert logger and isinstance(logger, Logger)
    assert request and request.messages

    logger.debug(
        "Processing chat completion request with messages %r", request
    )

    messages = [
        Message(role=req.role, content=_to_message_content(req.content))
        for req in request.messages
    ]

    logger.debug(
        "Transformed chat completion request to engine input %r", messages
    )

    response_id = uuid4()
    timestamp = int(time())

    settings = GenerationSettings(
        temperature=request.temperature,
        max_new_tokens=request.max_tokens,
        stop_strings=request.stop,
        top_p=request.top_p,
        # num_return_sequences=request.n
    )

    logger.debug(
        "Calling orchestrator with input %r and settings %r for response %s",
        messages,
        settings,
        response_id,
    )

    response = await orchestrator(messages, settings=settings)

    # Streaming through SSE (server-sent events with text/event-stream)
    if request.stream:

        async def generate_chunks():
            async for token in response:
                if isinstance(token, Event):
                    continue

                choice = ChatCompletionChunkChoice(
                    delta=ChatCompletionChunkChoiceDelta(content=token)
                )
                chunk = ChatCompletionChunk(
                    id=str(response_id),
                    created=timestamp,
                    model=request.model,
                    choices=[choice],
                )
                yield f"data: {chunk.model_dump_json()}\n\n"  # SSE data event
            yield "data: [DONE]\n\n"  # end of stream

        logger.debug(
            f"Generating event-stream stream for response {response_id}"
        )

        return StreamingResponse(
            generate_chunks(), media_type="text/event-stream"
        )

    # Non streaming
    message = ChatMessage(
        role=str(MessageRole.ASSISTANT), content=await response.to_str()
    )
    usage = ChatCompletionUsage()
    response = ChatCompletionResponse(
        id=str(response_id),
        created=timestamp,
        model=request.model,
        choices=[ChatCompletionChoice(message=message, finish_reason="stop")],
        usage=usage,
    )
    logger.debug(
        "Generated chat completion response #%s %r", response_id, response
    )
    return response


def _to_message_content(item):
    if isinstance(item, list):
        return [
            _to_message_content(i)
            for i in item
            if isinstance(i, (ContentImage, ContentText, str))
        ]
    if isinstance(item, ContentImage):
        return MessageContentImage(type=item.type, image_url=item.image_url)
    if isinstance(item, ContentText):
        return MessageContentText(type=item.type, text=item.text)
    if isinstance(item, str):
        return MessageContentText(type="text", text=item)
    raise TypeError(f"Unsupported content type: {type(item).__name__}")



from abc import ABC, abstractmethod
from ..entities import EngineMessage
from dataclasses import dataclass
from threading import Lock
from typing import Generic, TypeVar
from ..compat import override
from uuid import UUID

T = TypeVar("T")


@dataclass(frozen=True)
class MemoryChunk(Generic[T]):
    repository_key: str
    key: str | None
    data: T


class MemoryStore(ABC, Generic[T]):
    @abstractmethod
    async def append(self, agent_id: UUID, data: T) -> None:
        raise NotImplementedError()

    @abstractmethod
    async def reset(self) -> None:
        raise NotImplementedError()

    @abstractmethod
    async def search(self, query: str) -> list[T] | None:
        raise NotImplementedError()


class MessageMemory(MemoryStore[EngineMessage], ABC):
    def search(self, query: str) -> list[EngineMessage] | None:
        raise NotImplementedError()


class RecentMessageMemory(MessageMemory):
    _lock: Lock
    _data: list[EngineMessage]

    def __init__(self, **kwargs) -> None:
        self._lock = Lock()
        self.reset()
        super().__init__(**kwargs)

    @override
    def append(self, data: EngineMessage) -> None:
        with self._lock:
            self._data.append(data)

    def reset(self) -> None:
        with self._lock:
            self._data = []

    @property
    def size(self) -> int:
        return len(self._data)

    @property
    def is_empty(self) -> bool:
        return not bool(self._data)

    @property
    def data(self) -> list[EngineMessage]:
        return self._data



from abc import ABC, abstractmethod
from ...model import TokenizerNotSupportedException
from ...model.engine import Engine
from PIL import Image
from torch import Tensor
from torchaudio import load
from torchaudio.functional import resample
from numpy import ndarray
from transformers import (
    PreTrainedTokenizer,
    PreTrainedTokenizerFast,
)
from typing import Literal


class BaseAudioModel(Engine, ABC):
    @abstractmethod
    async def __call__(
        self,
        image_source: str | Image.Image,
        tensor_format: Literal["pt"] = "pt",
    ) -> str:
        raise NotImplementedError()

    def _load_tokenizer(
        self, tokenizer_name_or_path: str | None, use_fast: bool = True
    ) -> PreTrainedTokenizer | PreTrainedTokenizerFast:
        raise TokenizerNotSupportedException()

    def _load_tokenizer_with_tokens(
        self, tokenizer_name_or_path: str | None, use_fast: bool = True
    ) -> PreTrainedTokenizer | PreTrainedTokenizerFast:
        raise TokenizerNotSupportedException()

    def _resample_mono(self, audio_source: str, sampling_rate: int) -> Tensor:
        wave, wave_sampling_rate = load(audio_source)

        if wave.shape[0] > 1:
            # stereo -> mono
            wave = wave.mean(dim=0)
        else:
            # already mono, just drop channel dim (samples,)
            wave = wave.squeeze(0)

        if wave_sampling_rate != sampling_rate:
            wave = resample(
                wave.unsqueeze(0), wave_sampling_rate, sampling_rate
            ).squeeze(0)

        return wave

    def _resample(self, audio_source: str, sampling_rate: int) -> ndarray:
        wave, wave_sampling_rate = load(audio_source)
        if wave_sampling_rate != sampling_rate:
            wave = resample(wave, wave_sampling_rate, sampling_rate)
        wave = wave.mean(0).numpy()
        return wave



from ...compat import override
from ...model.vendor import TextGenerationVendor
from ...model.vision import BaseVisionModel
from ...model.engine import Engine
from diffusers import DiffusionPipeline
from PIL import Image
from torch import inference_mode, unique
from transformers import (
    AutoImageProcessor,
    AutoModelForSemanticSegmentation,
    PreTrainedModel,
)
from typing import Literal


class SemanticSegmentationModel(BaseVisionModel):
    def _load_model(
        self,
    ) -> PreTrainedModel | TextGenerationVendor | DiffusionPipeline:
        self._processor = AutoImageProcessor.from_pretrained(
            self._model_id,
            # default behavior in transformers v4.48
            use_fast=True,
        )
        model = AutoModelForSemanticSegmentation.from_pretrained(
            self._model_id,
            device_map=self._device,
            tp_plan=Engine._get_tp_plan(self._settings.parallel),
        )
        model.eval()
        return model

    @override
    async def __call__(
        self,
        image_source: str | Image.Image,
        tensor_format: Literal["pt"] = "pt",
    ) -> list[str]:
        image = BaseVisionModel._get_image(image_source)
        inputs = self._processor(images=image, return_tensors=tensor_format)
        inputs.to(self._device)
        with inference_mode():
            logits = self._model(**inputs).logits
        # shape (height, width) with class indices
        mask = logits.argmax(dim=1)[0]
        labels_tensor = unique(mask)
        labels = [
            self._model.config.id2label[idx.item()] for idx in labels_tensor
        ]
        return labels



from ...compat import override
from ...filters import Partitioner
from ...entities import TextPartition
from ...model.nlp.sentence import SentenceTransformerModel
from logging import Logger
from re import split
from transformers import PreTrainedTokenizer, PreTrainedTokenizerFast
from typing import Callable


class TextPartitioner(Partitioner):
    _PARAGRAPH_PATTERN = r"(?:\r\n|\r|\n){2,}"
    _model: SentenceTransformerModel
    _tokenizer: PreTrainedTokenizer | PreTrainedTokenizerFast
    _logger: Logger
    _max_tokens: int
    _window_size: int
    _overlap_size: int

    def __init__(
        self,
        model: SentenceTransformerModel,
        logger: Logger,
        *args,
        max_tokens: int,
        overlap_size: int,
        window_size: int,
    ):
        assert model and model.tokenizer
        self._model = model
        self._tokenizer = model.tokenizer
        self._logger = logger
        self.configure(
            max_tokens=max_tokens,
            overlap_size=overlap_size,
            window_size=window_size,
        )

    def configure(
        self,
        *args,
        max_tokens: int,
        overlap_size: int,
        window_size: int,
    ) -> None:
        assert (
            max_tokens
            and window_size
            and overlap_size
            and max_tokens > 0
            and window_size > 0
            and overlap_size > 0
            and window_size < max_tokens
            and overlap_size < max_tokens
            and window_size > overlap_size
        )
        self._max_tokens = max_tokens
        self._window_size = window_size
        self._overlap_size = overlap_size

    @override
    @property
    def sentence_model(self) -> Callable:
        return self._model

    @override
    async def __call__(
        self,
        text: str,
    ) -> list[TextPartition]:
        assert text
        partitions: list[TextPartition] = []

        for paragraph in split(self._PARAGRAPH_PATTERN, text):
            paragraph = paragraph.strip()
            if not paragraph:
                continue
            token_ids = self._tokenizer.encode(
                paragraph,
                add_special_tokens=False,
            )
            length = len(token_ids)
            if length <= self._max_tokens:
                embeddings = await self._model(paragraph)
                partitions.append(
                    TextPartition(
                        data=paragraph,
                        embeddings=embeddings,
                        total_tokens=length,
                    )
                )
            else:
                step = self._window_size - self._overlap_size

                for start in range(0, length, step):
                    window_token_ids = token_ids[
                        start : start + self._window_size
                    ]
                    section_length = len(window_token_ids)
                    section = self._tokenizer.decode(
                        window_token_ids, skip_special_tokens=True
                    )
                    embeddings = await self._model(section)
                    partitions.append(
                        TextPartition(
                            data=section,
                            embeddings=embeddings,
                            total_tokens=section_length,
                        )
                    )

        return partitions



from ...agent.loader import OrchestratorLoader
from ...agent.orchestrator import Orchestrator
from ...agent.orchestrator.response.orchestrator_response import (
    OrchestratorResponse,
)
from ...cli import get_input, has_input, confirm_tool_call
from ...cli.commands.model import token_generation
from ...entities import Backend, OrchestratorSettings, ToolCall
from ...event import EventStats
from ...model.hubs.huggingface import HuggingfaceHub
from ...model.nlp.text.vendor import TextGenerationVendorModel
from ...server import agents_server
from ...tool.browser import BrowserToolSettings
from argparse import Namespace
from contextlib import AsyncExitStack
from dataclasses import fields
from jinja2 import Environment, FileSystemLoader
from logging import Logger
from os.path import dirname, join, getmtime
from rich.console import Console
from rich.prompt import Confirm, Prompt
from rich.syntax import Syntax
from rich.theme import Theme
from typing import Mapping
from uuid import UUID, uuid4


def get_orchestrator_settings(
    args: Namespace,
    *,
    agent_id: UUID,
    name: str | None = None,
    role: str | None = None,
    task: str | None = None,
    instructions: str | None = None,
    engine_uri: str | None = None,
    memory_recent: bool | None = None,
    memory_permanent_message: str | None = None,
    memory_permanent: list[str] | None = None,
    max_new_tokens: int | None = None,
    temperature: float | None = None,
    tools: list[str] | None = None,
    top_k: int | None = None,
    top_p: float | None = None,
) -> OrchestratorSettings:
    """Create ``OrchestratorSettings`` from CLI arguments."""
    memory_recent = (
        memory_recent
        if memory_recent is not None
        else (
            args.memory_recent
            if args.memory_recent is not None
            else not getattr(args, "no_session", False)
        )
    )
    engine_uri = engine_uri or args.engine_uri
    call_tokens = (
        max_new_tokens
        if max_new_tokens is not None
        else args.run_max_new_tokens
    )

    chat_settings = {
        k[len("run_chat_") :]: v
        for k, v in vars(args).items()
        if k.startswith("run_chat_") and v is not None
    }

    return OrchestratorSettings(
        agent_id=agent_id,
        orchestrator_type=None,
        agent_config={
            k: v
            for k, v in {
                "name": name if name is not None else args.name,
                "role": role if role is not None else args.role,
                "task": task if task is not None else args.task,
                "instructions": (
                    instructions
                    if instructions is not None
                    else args.instructions
                ),
            }.items()
            if v is not None
        },
        uri=engine_uri,
        engine_config={
            "backend": getattr(args, "backend", Backend.TRANSFORMERS.value)
        },
        call_options={
            "max_new_tokens": call_tokens,
            "skip_special_tokens": args.run_skip_special_tokens,
            "temperature": temperature,
            "top_k": top_k,
            "top_p": top_p,
            **({"chat_settings": chat_settings} if chat_settings else {}),
        },
        template_vars=None,
        memory_permanent_message=(
            memory_permanent_message
            if memory_permanent_message is not None
            else args.memory_permanent_message
        ),
        permanent_memory=(
            {
                ns: dsn
                for ns, dsn in (
                    item.split("@", 1)
                    for item in (
                        memory_permanent or args.memory_permanent or []
                    )
                )
            }
            if (memory_permanent or args.memory_permanent)
            else None
        ),
        memory_recent=memory_recent,
        sentence_model_id=(
            args.memory_engine_model_id
            or OrchestratorLoader.DEFAULT_SENTENCE_MODEL_ID
        ),
        sentence_model_engine_config=None,
        sentence_model_max_tokens=args.memory_engine_max_tokens,
        sentence_model_overlap_size=args.memory_engine_overlap,
        sentence_model_window_size=args.memory_engine_window,
        json_config=None,
        tools=tools if tools is not None else args.tool or [],
        log_events=True,
    )


def _tool_settings_from_mapping(
    mapping: Mapping[str, object] | Namespace,
    *,
    prefix: str | None = None,
    settings_cls: type,
    open_files: bool = True,
) -> object:
    """Return tool settings from a mapping using dataclass ``settings_cls``."""
    values: dict[str, object] = {}
    for field in fields(settings_cls):
        key = f"tool_{prefix}_{field.name}" if prefix else field.name
        if isinstance(mapping, Namespace):
            if hasattr(mapping, key):
                value = getattr(mapping, key)
            else:
                continue
        else:
            if key in mapping:
                value = mapping[key]
            elif prefix and field.name in mapping:
                value = mapping[field.name]
            else:
                continue

        if value is not None:
            if (
                field.name == "debug_source"
                and open_files
                and isinstance(value, str)
            ):
                value = open(value)
            values[field.name] = value

    if not values:
        return None

    return settings_cls(**values)


def get_tool_settings(
    args: Namespace,
    *,
    prefix: str,
    settings_cls: type,
    open_files: bool = True,
) -> object:
    return _tool_settings_from_mapping(
        args, prefix=prefix, settings_cls=settings_cls, open_files=open_files
    )


async def agent_message_search(
    args: Namespace,
    console: Console,
    theme: Theme,
    hub: HuggingfaceHub,
    logger: Logger,
    refresh_per_second: int,
) -> None:
    _, _i = theme._, theme.icons

    specs_path = args.specifications_file
    engine_uri = getattr(args, "engine_uri", None)
    assert not (
        specs_path and engine_uri
    ), "specifications file and --engine-uri are mutually exclusive"
    assert (
        specs_path or engine_uri
    ), "specifications file or --engine-uri must be specified"
    agent_id = args.id
    participant_id = args.participant
    session_id = args.session

    assert agent_id and participant_id and session_id

    input_string = get_input(
        console,
        _i["user_input"] + " ",
        echo_stdin=not args.no_repl,
        is_quiet=args.quiet,
    )
    if not input_string:
        return

    limit = args.limit

    async with AsyncExitStack() as stack:
        loader = OrchestratorLoader(
            hub=hub,
            logger=logger,
            participant_id=participant_id,
            stack=stack,
        )
        with console.status(
            _("Loading agent..."),
            spinner=theme.get_spinner("agent_loading"),
            refresh_per_second=refresh_per_second,
        ):
            if specs_path:
                logger.debug(
                    "Loading agent from %s for participant %s",
                    specs_path,
                    participant_id,
                )

                orchestrator = await loader.from_file(
                    specs_path,
                    agent_id=agent_id,
                )
            else:
                assert (
                    args.engine_uri
                ), "--engine-uri required when no specifications file"
                logger.debug("Loading agent from inline settings")
                memory_recent = (
                    args.memory_recent
                    if args.memory_recent is not None
                    else True
                )
                settings = get_orchestrator_settings(
                    args,
                    agent_id=agent_id,
                    memory_recent=memory_recent,
                    tools=args.tool,
                )
                browser_settings = get_tool_settings(
                    args, prefix="browser", settings_cls=BrowserToolSettings
                )
                orchestrator = await loader.from_settings(
                    settings,
                    browser_settings=browser_settings,
                )
            orchestrator = await stack.enter_async_context(orchestrator)

            assert orchestrator.engine_agent and orchestrator.engine.model_id

            can_access = args.skip_hub_access_check or hub.can_access(
                orchestrator.engine.model_id
            )
            is_local = not isinstance(
                orchestrator.engine, TextGenerationVendorModel
            )
            models = [
                hub.model(model_id) if is_local else model_id
                for model_id in orchestrator.model_ids
            ]

            console.print(
                theme.agent(orchestrator, models=models, can_access=can_access)
            )

            logger.debug(
                'Searching for "%s" across messages on session %s between '
                "agent %s and participant %s",
                input_string,
                session_id,
                agent_id,
                participant_id,
            )
            messages = await orchestrator.memory.search_messages(
                search=input_string,
                agent_id=agent_id,
                search_user_messages=False,
                session_id=session_id,
                participant_id=participant_id,
                function=args.function,
                limit=limit,
            )
            console.print(
                theme.search_message_matches(
                    participant_id, orchestrator, messages
                )
            )


async def agent_run(
    args: Namespace,
    console: Console,
    theme: Theme,
    hub: HuggingfaceHub,
    logger: Logger,
    refresh_per_second: int,
) -> None:
    _, _i = theme._, theme.icons

    specs_path = args.specifications_file
    engine_uri = getattr(args, "engine_uri", None)
    assert not (
        specs_path and engine_uri
    ), "specifications file and --engine-uri are mutually exclusive"
    assert (
        specs_path or engine_uri
    ), "specifications file or --engine-uri must be specified"
    use_async_generator = not args.use_sync_generator
    display_tokens = args.display_tokens or 0
    dtokens_pick = 10 if display_tokens > 0 else 0
    with_stats = args.stats and not args.quiet
    agent_id = args.id
    participant_id = args.participant
    session_id = args.session if not args.no_session else None
    load_recent_messages = (
        not args.skip_load_recent_messages and not args.no_session
    )
    load_recent_messages_limit = args.load_recent_messages_limit

    event_stats = EventStats()

    def _confirm_call(call: ToolCall) -> str:
        return confirm_tool_call(console, call)

    async def _event_listener(event):
        nonlocal event_stats
        event_stats.total_triggers += 1
        if event.type not in event_stats.triggers:
            event_stats.triggers[event.type] = 1
        else:
            event_stats.triggers[event.type] += 1

    async def _init_orchestrator() -> Orchestrator:
        loader = OrchestratorLoader(
            hub=hub,
            logger=logger,
            participant_id=participant_id,
            stack=stack,
        )
        if specs_path:
            logger.debug(
                "Loading agent from %s for participant %s",
                specs_path,
                participant_id,
            )

            orchestrator = await loader.from_file(
                specs_path,
                agent_id=agent_id,
                disable_memory=args.no_session,
            )
        else:
            assert (
                args.engine_uri and args.role
            ), "--engine-uri and --role required when no specifications file"
            assert not args.specifications_file or not args.engine_uri
            memory_recent = (
                args.memory_recent
                if args.memory_recent is not None
                else not args.no_session
            )
            settings = get_orchestrator_settings(
                args,
                agent_id=agent_id or uuid4(),
                memory_recent=memory_recent,
                tools=args.tool,
            )
            logger.debug("Loading agent from inline settings")
            browser_settings = get_tool_settings(
                args, prefix="browser", settings_cls=BrowserToolSettings
            )
            orchestrator = await loader.from_settings(
                settings,
                browser_settings=browser_settings,
            )
        orchestrator.event_manager.add_listener(_event_listener)

        orchestrator = await stack.enter_async_context(orchestrator)

        if args.tools_confirm:
            assert (
                not orchestrator.tool.is_empty
            ), "--tools-confirm requires tools"

        logger.debug(
            "Agent loaded from %s, models used: %s, with recent message "
            "memory: %s, with permanent message memory: %s",
            specs_path,
            orchestrator.model_ids,
            "yes" if orchestrator.memory.has_recent_message else "no",
            (
                "yes, with session #"
                + str(orchestrator.memory.permanent_message.session_id)
                if orchestrator.memory.has_permanent_message
                else "no"
            ),
        )

        if not args.quiet:
            assert orchestrator.engine_agent and orchestrator.engine.model_id

            is_local = not isinstance(
                orchestrator.engine, TextGenerationVendorModel
            )

            can_access = (
                args.skip_hub_access_check
                or not is_local
                or hub.can_access(orchestrator.engine.model_id)
            )
            models = [
                hub.model(model_id) if is_local else model_id
                for model_id in orchestrator.model_ids
            ]

            console.print(
                theme.agent(orchestrator, models=models, can_access=can_access)
            )

        if not args.no_session:
            if session_id:
                await orchestrator.memory.continue_session(
                    session_id=session_id,
                    load_recent_messages=load_recent_messages,
                    load_recent_messages_limit=load_recent_messages_limit,
                )
            else:
                await orchestrator.memory.start_session()

        if (
            load_recent_messages
            and orchestrator.memory.has_recent_message
            and not orchestrator.memory.recent_message.is_empty
            and not args.quiet
        ):
            console.print(
                theme.recent_messages(
                    participant_id,
                    orchestrator,
                    orchestrator.memory.recent_message.data,
                )
            )

        return orchestrator

    async with AsyncExitStack() as stack:
        with console.status(
            _("Loading agent..."),
            spinner=theme.get_spinner("agent_loading"),
            refresh_per_second=refresh_per_second,
        ):
            orchestrator = await _init_orchestrator()

        watch_spec = bool(specs_path and args.conversation and args.watch)
        if watch_spec:
            specs_mtime = getmtime(specs_path)

        input_string: str | None = None
        in_conversation = False
        while not input_string or in_conversation:
            if watch_spec and not has_input(console):
                new_mtime = getmtime(specs_path)
                if new_mtime != specs_mtime:
                    logger.debug("Reloading agent from %s", specs_path)
                    orchestrator = await _init_orchestrator()
                    specs_mtime = new_mtime
                    in_conversation = False
                    continue
            logger.debug(
                "Waiting for new message to add to orchestrator's existing "
                + str(orchestrator.memory.recent_message.size)
                if orchestrator.memory
                and orchestrator.memory.has_recent_message
                else "0" + " messages"
            )
            input_string = get_input(
                console,
                _i["user_input"] + " ",
                echo_stdin=not args.no_repl,
                force_prompt=in_conversation,
                is_quiet=args.quiet,
                tty_path=args.tty,
            )
            if not input_string:
                logger.debug("Finishing session with orchestrator")
                return

            logger.debug('Agent about to process input "%s"', input_string)
            output = await orchestrator(
                input_string,
                use_async_generator=use_async_generator,
                tool_confirm=_confirm_call if args.tools_confirm else None,
            )

            if not args.quiet and not args.stats:
                console.print(_i["agent_output"] + " ", end="")

            if args.quiet:
                console.print(await output.to_str())
                return

            assert isinstance(output, OrchestratorResponse)

            await token_generation(
                args=args,
                console=console,
                theme=theme,
                logger=logger,
                orchestrator=orchestrator,
                event_stats=event_stats,
                lm=orchestrator.engine,
                input_string=input_string,
                refresh_per_second=refresh_per_second,
                response=output,
                dtokens_pick=dtokens_pick,
                display_tokens=display_tokens,
                tool_events_limit=args.display_tools_events,
                with_stats=with_stats,
            )

            if args.conversation:
                console.print("")
                if not in_conversation:
                    in_conversation = True
            else:
                break


async def agent_serve(
    args: Namespace,
    hub: HuggingfaceHub,
    logger: Logger,
    name: str,
    version: str,
) -> None:
    assert args.host and args.port
    specs_path = args.specifications_file
    engine_uri = getattr(args, "engine_uri", None)
    assert not (
        specs_path and engine_uri
    ), "specifications file and --engine-uri are mutually exclusive"
    assert (
        specs_path or engine_uri
    ), "specifications file or --engine-uri must be specified"

    async with AsyncExitStack() as stack:
        loader = OrchestratorLoader(
            hub=hub,
            logger=logger,
            participant_id=uuid4(),
            stack=stack,
        )
        if specs_path:
            logger.debug("Loading agent from %s", specs_path)

            orchestrator = await loader.from_file(
                specs_path,
                agent_id=uuid4(),
            )
        else:
            assert args.role
            memory_recent = (
                args.memory_recent if args.memory_recent is not None else True
            )
            settings = get_orchestrator_settings(
                args,
                agent_id=uuid4(),
                memory_recent=memory_recent,
                tools=args.tool,
            )
            logger.debug("Loading agent from inline settings")
            browser_settings = get_tool_settings(
                args, prefix="browser", settings_cls=BrowserToolSettings
            )
            orchestrator = await loader.from_settings(
                settings,
                browser_settings=browser_settings,
            )
        orchestrator = await stack.enter_async_context(orchestrator)

        logger.debug(
            "Agent loaded from"
            f" {specs_path if specs_path else 'inline settings'}"
        )
        server = agents_server(
            name=name,
            version=version,
            prefix_openai=args.prefix_openai,
            prefix_mcp=args.prefix_mcp,
            orchestrators=[orchestrator],
            host=args.host,
            port=args.port,
            reload=args.reload,
            logger=logger,
        )
        await server.serve()


async def agent_init(args: Namespace, console: Console, theme: Theme) -> None:
    _ = theme._

    name = args.name or Prompt.ask(_("Agent name"))
    role = args.role or get_input(
        console,
        _("Agent role") + " ",
        echo_stdin=not args.no_repl,
        is_quiet=args.quiet,
    )
    if not role:
        return

    task = args.task or get_input(
        console,
        _("Agent task") + " ",
        echo_stdin=not args.no_repl,
        is_quiet=args.quiet,
    )
    instructions = args.instructions or get_input(
        console,
        _("Agent instructions") + " ",
        echo_stdin=not args.no_repl,
        is_quiet=args.quiet,
    )

    memory_recent = (
        args.memory_recent
        if args.memory_recent is not None
        else Confirm.ask(_("Use recent message memory?"))
    )
    memory_permanent_message = (
        args.memory_permanent_message
        if args.memory_permanent_message is not None
        else Prompt.ask(_("Permanent memory DSN"), default="")
    )
    engine_uri = args.engine_uri or Prompt.ask(
        _("Engine URI"),
        default="microsoft/Phi-4-mini-instruct",
    )

    settings = get_orchestrator_settings(
        args,
        agent_id=uuid4(),
        name=name,
        role=role,
        task=task,
        instructions=instructions,
        engine_uri=engine_uri,
        memory_recent=memory_recent,
        memory_permanent_message=memory_permanent_message,
        max_new_tokens=args.run_max_new_tokens or 1024,
        tools=args.tool or [],
    )

    browser_tool = get_tool_settings(
        args,
        prefix="browser",
        settings_cls=BrowserToolSettings,
        open_files=False,
    )

    env = Environment(
        loader=FileSystemLoader(join(dirname(__file__), "..", "..", "agent")),
        trim_blocks=True,
        lstrip_blocks=True,
    )
    template = env.get_template("blueprint.toml")
    rendered = template.render(
        orchestrator=settings, browser_tool=browser_tool
    )
    console.print(Syntax(rendered, "toml"))



from ..event import Event, EventType
from asyncio import Event as EventSignal, Queue, TimeoutError, wait_for
from collections import defaultdict, deque
from inspect import iscoroutine
from typing import Awaitable, Callable, Iterable

Listener = Callable[[Event], Awaitable[None] | None]


class EventManager:
    _listeners: dict[EventType, list[Listener]]
    _queue: Queue[Event]
    _history: deque[Event]

    def __init__(self, history_length: int | None = None) -> None:
        self._listeners = defaultdict(list)
        self._queue = Queue()
        self._history = deque(maxlen=history_length)

    @property
    def history(self) -> list[Event]:
        return list(self._history)

    def add_listener(
        self,
        listener: Listener,
        event_types: Iterable[EventType] | None = None,
    ) -> None:
        types = list(event_types) if event_types else list(EventType)
        for event_type in types:
            self._listeners[event_type].append(listener)

    async def trigger(self, event: Event) -> None:
        self._history.append(event)
        self._queue.put_nowait(event)

        for listener in self._listeners.get(event.type, []):
            result = listener(event)
            if iscoroutine(result):
                await result

    async def listen(self, stop_signal: EventSignal, timeout: float = 0.2):
        while True:
            try:
                evt = await wait_for(self._queue.get(), timeout=timeout)
                yield evt
            except TimeoutError:
                if stop_signal.is_set() and self._queue.empty():
                    break



from . import Secrets
from keyring import get_keyring
from keyring.backend import KeyringBackend


class KeyringSecrets(Secrets):
    _SERVICE = "avalan"

    def __init__(self, ring: KeyringBackend | None = None):
        self._ring = get_keyring()

    def read(self, key: str) -> str | None:
        return self._ring.get_password(self._SERVICE, key)

    def write(self, key: str, secret: str) -> None:
        self._ring.set_password(self._SERVICE, key, secret)

    def delete(self, key: str) -> None:
        self._ring.delete_password(self._SERVICE, key)



from .entities import TextPartition
from abc import ABC, abstractmethod
from typing import Callable


class Partitioner(ABC):
    @property
    @abstractmethod
    def sentence_model(self) -> Callable:
        raise NotImplementedError()

    @abstractmethod
    async def __call__(
        self,
        text: str,
    ) -> list[TextPartition]:
        raise NotImplementedError()



from . import Tool, ToolSet
from ..compat import override
from ..entities import ToolCallContext
from contextlib import AsyncExitStack
from RestrictedPython import (
    compile_restricted,
    safe_globals,
    RestrictingNodeTransformer,
)


class CodeTool(Tool):
    """Execute Python function code in a restricted environment.

    Args:
        code: Python code to execute, which should define a function named
        `run`, with whatever arguments needed, and returning its result on
        a string.
        args: Positional arguments to send the `run` function.
        kwargs: Keyword arguments to send the `run` function.

    Returns:
        A string, which is the value returned by the `run` function.
    """

    def __init__(self) -> None:
        super().__init__()
        self.__name__ = "run"

    async def __call__(
        self, code: str, *args: any, context: ToolCallContext, **kwargs: any
    ) -> str:
        locals_dict = {}
        byte_code = compile_restricted(
            code,
            filename="<avalan:tool:code>",
            mode="exec",
            flags=0,
            dont_inherit=False,
            policy=RestrictingNodeTransformer,
        )
        exec(byte_code, safe_globals, locals_dict)
        assert "run" in locals_dict

        function = locals_dict["run"]

        if args and not kwargs and isinstance(args, tuple) and len(args) == 2:
            (args, kwargs) = args
            if args and not kwargs and isinstance(args, dict):
                kwargs = args
                args = None

        result = (
            function(*args, **kwargs)
            if args and kwargs
            else (
                function(*args)
                if args
                else function(**kwargs) if kwargs else function()
            )
        )

        return str(result)


class CodeToolSet(ToolSet):
    @override
    def __init__(
        self,
        *,
        exit_stack: AsyncExitStack | None = None,
        namespace: str | None = None,
    ) -> None:
        tools = [CodeTool()]
        super().__init__(
            exit_stack=exit_stack, namespace=namespace, tools=tools
        )



from datetime import datetime, timezone
from json import dumps, loads
from logging import Logger
from typing import Any
from uuid import UUID, uuid4

from boto3 import client as boto_client
from ....deploy.aws import AsyncClient

from ....entities import (
    EngineMessage,
    EngineMessageScored,
    Message,
    MessageRole,
)
from ....memory.partitioner.text import TextPartition
from ....memory.permanent import (
    PermanentMessageMemory,
    VectorFunction,
)
from . import S3VectorsMemory


class S3VectorsMessageMemory(S3VectorsMemory, PermanentMessageMemory):
    def __init__(
        self,
        bucket: str,
        collection: str,
        *,
        client: Any,
        logger: Logger,
        sentence_model: Any | None = None,
    ) -> None:
        S3VectorsMemory.__init__(
            self,
            bucket=bucket,
            collection=collection,
            client=client,
            logger=logger,
        )
        PermanentMessageMemory.__init__(self, sentence_model=sentence_model)  # type: ignore[arg-type]

    @classmethod
    async def create_instance(
        cls,
        bucket: str,
        collection: str,
        *,
        logger: Logger,
        aws_client: Any | None = None,
        sentence_model: Any | None = None,
    ) -> "S3VectorsMessageMemory":
        if aws_client is None:
            aws_client = boto_client("s3vectors")
        return cls(
            bucket=bucket,
            collection=collection,
            client=AsyncClient(aws_client),
            logger=logger,
            sentence_model=sentence_model,
        )

    async def create_session(
        self, *, agent_id: UUID, participant_id: UUID
    ) -> UUID:
        return uuid4()

    async def continue_session_and_get_id(
        self,
        *,
        agent_id: UUID,
        participant_id: UUID,
        session_id: UUID,
    ) -> UUID:
        return session_id

    async def append_with_partitions(
        self,
        engine_message: EngineMessage,
        *,
        partitions: list[TextPartition],
    ) -> None:
        assert engine_message and partitions
        now_utc = datetime.now(timezone.utc)
        message, message_partitions = self._build_message_with_partitions(
            engine_message,
            self._session_id,
            partitions,
            created_at=now_utc,
            message_id=uuid4(),
        )
        key = (
            f"{self._collection}/{message.session_id}/{message.id}.json"
            if message.session_id
            else f"{self._collection}/{message.id}.json"
        )
        await self._put_object(
            Bucket=self._bucket,
            Key=key,
            Body=dumps(
                {
                    "id": str(message.id),
                    "agent_id": str(message.agent_id),
                    "model_id": message.model_id,
                    "session_id": (
                        str(message.session_id) if message.session_id else None
                    ),
                    "author": str(message.author),
                    "data": message.data,
                    "partitions": message.partitions,
                    "created_at": message.created_at.isoformat(),
                }
            ).encode(),
        )
        for row in message_partitions:
            await self._put_vector(
                Bucket=self._bucket,
                Collection=self._collection,
                Id=f"{row.message_id}:{row.partition}",
                Vector=row.embedding.tolist(),
                Metadata={
                    "message_id": str(row.message_id),
                    "agent_id": str(row.agent_id),
                    "session_id": (
                        str(row.session_id) if row.session_id else None
                    ),
                },
            )

    async def get_recent_messages(
        self,
        session_id: UUID,
        participant_id: UUID,
        *,
        limit: int | None = None,
    ) -> list[EngineMessage]:
        prefix = f"{self._collection}/{session_id}/"
        response = await self._call_client(
            self._client.list_objects_v2, Bucket=self._bucket, Prefix=prefix
        )
        objects = response.get("Contents", []) if response else []
        objs = sorted(
            objects, key=lambda o: o.get("LastModified"), reverse=True
        )
        messages: list[EngineMessage] = []
        for obj in objs[: limit or len(objs)]:
            data = await self._get_object(Bucket=self._bucket, Key=obj["Key"])
            meta = loads(data["Body"].read().decode())
            messages.append(
                EngineMessage(
                    agent_id=UUID(meta["agent_id"]),
                    model_id=meta["model_id"],
                    message=Message(
                        role=MessageRole(meta["author"]), content=meta["data"]
                    ),
                )
            )
        return messages

    async def search_messages(
        self,
        *,
        agent_id: UUID,
        function: VectorFunction,
        limit: int | None = None,
        participant_id: UUID,
        search_partitions: list[TextPartition],
        search_user_messages: bool,
        session_id: UUID | None,
        exclude_session_id: UUID | None,
    ) -> list[EngineMessageScored]:
        assert agent_id and participant_id and search_partitions
        query = search_partitions[0].embeddings.tolist()
        filt = {
            "message_id": "*",
            "agent_id": str(agent_id),
            "participant_id": str(participant_id),
        }
        if session_id:
            filt["session_id"] = str(session_id)
        response = await self._query_vector(
            Bucket=self._bucket,
            Collection=self._collection,
            QueryVector=query,
            TopK=limit or 10,
            Function=str(function),
            Filter=filt,
        )
        results: list[EngineMessageScored] = []
        for item in response.get("Items", []):
            msg_id = item.get("Metadata", {}).get("message_id")
            if not msg_id:
                continue
            key = (
                f"{self._collection}/{session_id}/{msg_id}.json"
                if session_id
                else f"{self._collection}/{msg_id}.json"
            )
            obj = await self._get_object(Bucket=self._bucket, Key=key)
            meta = loads(obj["Body"].read().decode())
            results.append(
                EngineMessageScored(
                    agent_id=UUID(meta["agent_id"]),
                    model_id=meta["model_id"],
                    message=Message(
                        role=MessageRole(meta["author"]), content=meta["data"]
                    ),
                    score=item.get("Score", 0.0),
                )
            )
        return results



from ..agent import Role, Specification
from ..agent.engine import EngineAgent
from ..memory.manager import MemoryManager
from ..model.engine import Engine
from ..model.manager import ModelManager
from ..tool.manager import ToolManager
from ..entities import EngineUri
from ..event.manager import EventManager
from jinja2 import (
    Environment as TemplateEnvironment,
    FileSystemLoader,
    Template,
)
from os import linesep
from os.path import dirname, join
from typing import Any
from uuid import UUID


class Renderer:
    _TEMPLATES_DIR = "templates"
    _environment: TemplateEnvironment
    _clean_spaces: bool
    _templates: dict[str, Template] = {}

    def __init__(
        self, templates_path: str | None = None, clean_spaces: bool = True
    ):
        self._clean_spaces = clean_spaces
        self._environment = TemplateEnvironment(
            loader=FileSystemLoader(
                templates_path or join(dirname(__file__), self._TEMPLATES_DIR)
            ),
            trim_blocks=True,
            lstrip_blocks=True,
        )

    def __call__(self, template_id: str, **kwargs) -> str:
        if template_id not in self._templates:
            self._templates[template_id] = self._environment.get_template(
                template_id
            )

        output = self._templates[template_id].render(**kwargs)

        if self._clean_spaces:
            output = linesep.join(line.strip() for line in output.splitlines())
        return output

    def from_string(
        self,
        template: str,
        template_vars: dict | None = None,
        encoding: str = "utf-8",
    ) -> str:
        return (
            Template(template).render(**template_vars).encode(encoding)
            if template_vars
            else template
        )


class TemplateEngineAgent(EngineAgent):
    _renderer: Renderer

    def __init__(
        self,
        model: Engine,
        memory: MemoryManager,
        tool: ToolManager,
        event_manager: EventManager,
        model_manager: ModelManager,
        renderer: Renderer,
        engine_uri: EngineUri,
        *args,
        name: str | None = None,
        id: UUID | None = None,
    ):
        super().__init__(
            model,
            memory,
            tool,
            event_manager,
            model_manager,
            engine_uri,
            name=name,
            id=id,
        )
        self._renderer = renderer

    def _prepare_call(
        self, specification: Specification, input: str, **kwargs: Any
    ) -> Any:
        template_id = specification.template_id or "agent.md"
        template_vars = specification.template_vars or {}
        template_vars.setdefault("name", self._name)
        template_vars.setdefault(
            "roles",
            (
                (
                    [
                        self._renderer.from_string(persona, template_vars)
                        for persona in specification.role.persona
                    ]
                    if isinstance(specification.role, Role)
                    else (
                        [specification.role]
                        if isinstance(specification.role, str)
                        else specification.role
                    )
                )
                if template_vars
                else specification.role
            ),
        )
        template_vars.setdefault(
            "task",
            (
                self._renderer.from_string(
                    specification.goal.task, template_vars
                )
                if specification.goal and template_vars
                else specification.goal.task if specification.goal else None
            ),
        )
        template_vars.setdefault(
            "instructions",
            (
                [
                    self._renderer.from_string(instruction, template_vars)
                    for instruction in specification.goal.instructions
                ]
                if specification.goal and template_vars
                else (
                    specification.goal.instructions
                    if specification.goal
                    else None
                )
            ),
        )
        template_vars.setdefault(
            "rules",
            (
                [
                    self._renderer.from_string(rule, template_vars)
                    for rule in specification.rules
                ]
                if template_vars and specification.rules
                else specification.rules
            ),
        )
        system_prompt = self._renderer(template_id, **template_vars)

        kwargs.setdefault("settings", specification.settings)
        kwargs.setdefault("system_prompt", system_prompt)
        return kwargs



from ....compat import override
from ....entities import BetaSchedule, EngineSettings, Input, TimestepSpacing
from ....model.vendor import TextGenerationVendor
from ....model.engine import Engine
from ....model.vision import BaseVisionModel
from dataclasses import replace
from diffusers import (
    AnimateDiffPipeline,
    DiffusionPipeline,
    EulerDiscreteScheduler,
    MotionAdapter,
)
from diffusers.schedulers.scheduling_utils import SchedulerMixin
from diffusers.utils import export_to_gif
from huggingface_hub import hf_hub_download
from logging import Logger, getLogger
from safetensors.torch import load_file
from torch import inference_mode
from transformers import PreTrainedModel


class TextToAnimationModel(BaseVisionModel):
    _schedulers: dict[tuple[TimestepSpacing, BetaSchedule], SchedulerMixin] = (
        {}
    )

    def __init__(
        self,
        model_id: str,
        settings: EngineSettings | None = None,
        logger: Logger = getLogger(__name__),
    ):
        settings = settings or EngineSettings()
        assert settings.base_model_id and settings.checkpoint
        settings = replace(settings, enable_eval=False)
        super().__init__(model_id, settings, logger)

    def _load_model(
        self,
    ) -> PreTrainedModel | TextGenerationVendor | DiffusionPipeline:
        dtype = Engine.weight(self._settings.weight_type)
        adapter = MotionAdapter().to(self._device, dtype)
        adapter.load_state_dict(
            load_file(
                hf_hub_download(self._model_id, self._settings.checkpoint),
                device=self._device,
            )
        )
        pipe = AnimateDiffPipeline.from_pretrained(
            self._settings.base_model_id,
            motion_adapter=adapter,
            torch_dtype=dtype,
        ).to(self._device)

        return pipe

    @override
    async def __call__(
        self,
        input: Input,
        path: str,
        *,
        beta_schedule: BetaSchedule = "linear",
        guidance_scale: float = 1.0,
        steps: int = 4,
        timestep_spacing: TimestepSpacing = "trailing",
    ) -> str:
        assert steps and steps in [
            1,
            2,
            4,
            8,
        ], f"Invalid number of steps: {steps}, can only be 1, 2, 4, or 8"
        scheduler_settings = (timestep_spacing, beta_schedule)
        if scheduler_settings not in self._schedulers:
            scheduler = EulerDiscreteScheduler.from_config(
                self._model.scheduler.config,
                timestep_spacing=timestep_spacing,
                beta_schedule=beta_schedule,
            )
            self._schedulers[scheduler_settings] = scheduler
        else:
            scheduler = self._schedulers[scheduler_settings]

        self._model.scheduler = scheduler

        with inference_mode():
            output = self._model(
                prompt=input if isinstance(input, str) else str(input),
                guidance_scale=guidance_scale,
                num_inference_steps=steps,
            )

        export_to_gif(output.frames[0], path)

        return path



from ..entities import (
    AttentionImplementation,
    EngineUri,
    GenerationSettings,
    ChatSettings,
    ReasoningSettings,
    Input,
    Modality,
    Operation,
    OperationAudioParameters,
    OperationParameters,
    OperationTextParameters,
    OperationVisionParameters,
    ParallelStrategy,
    Backend,
    TextGenerationLoaderClass,
    TransformerEngineSettings,
    Vendor,
    WeightType,
)
from ..model.hubs.huggingface import HuggingfaceHub
from ..model.nlp.sentence import SentenceTransformerModel
from ..model.nlp.text.generation import TextGenerationModel
from ..model.nlp.text.mlxlm import MlxLmModel
from ..model.nlp.question import QuestionAnsweringModel
from ..model.nlp.sequence import (
    SequenceClassificationModel,
    SequenceToSequenceModel,
    TranslationModel,
)
from ..model.nlp.token import TokenClassificationModel
from ..model.audio.classification import AudioClassificationModel
from ..model.audio.generation import AudioGenerationModel
from ..model.audio.speech_recognition import SpeechRecognitionModel
from ..model.audio.speech import TextToSpeechModel
from ..model.criteria import KeywordStoppingCriteria
from ..model.vision.detection import ObjectDetectionModel
from ..model.vision.classification import ImageClassificationModel
from ..model.vision.decoder import VisionEncoderDecoderModel
from ..model.vision.diffusion import (
    TextToAnimationModel,
    TextToImageModel,
    TextToVideoModel,
)
from ..model.vision.segmentation import SemanticSegmentationModel
from ..model.vision.text import ImageTextToTextModel, ImageToTextModel
from ..secrets import KeyringSecrets
from ..tool.manager import ToolManager
from ..event import Event, EventType
from ..event.manager import EventManager
from argparse import Namespace
from contextlib import ContextDecorator, ExitStack
from logging import Logger
from typing import Any, get_args, TypeAlias
from time import perf_counter
from urllib.parse import urlparse, parse_qsl

ModelType: TypeAlias = (
    AudioClassificationModel
    | AudioGenerationModel
    | ImageClassificationModel
    | ImageTextToTextModel
    | ImageToTextModel
    | ObjectDetectionModel
    | QuestionAnsweringModel
    | SemanticSegmentationModel
    | SentenceTransformerModel
    | SequenceClassificationModel
    | SequenceToSequenceModel
    | SpeechRecognitionModel
    | AudioGenerationModel
    | TextGenerationModel
    | TextToImageModel
    | TextToSpeechModel
    | TranslationModel
    | TokenClassificationModel
    | VisionEncoderDecoderModel
    | TextToAnimationModel
    | TextToVideoModel
)


class ModelManager(ContextDecorator):
    _hub: HuggingfaceHub
    _stack: ExitStack
    _logger: Logger
    _secrets: KeyringSecrets
    _event_manager: EventManager | None

    def __init__(
        self,
        hub: HuggingfaceHub,
        logger: Logger,
        secrets: KeyringSecrets | None = None,
        event_manager: EventManager | None = None,
    ):
        self._hub, self._logger = hub, logger
        self._stack = ExitStack()
        self._secrets = secrets or KeyringSecrets()
        self._event_manager = event_manager

    def __enter__(self):
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: Any | None,
    ):
        return self._stack.__exit__(exc_type, exc_value, traceback)

    async def __call__(
        self,
        engine_uri: EngineUri,
        model: ModelType,
        operation: Operation,
        tool: ToolManager | None = None,
    ):
        modality = operation.modality

        stopping_criteria = (
            KeywordStoppingCriteria(
                operation.parameters["text"].stop_on_keywords,
                model.tokenizer,
            )
            if operation.parameters
            and "text" in operation.parameters
            and operation.parameters["text"]
            and operation.parameters["text"].stop_on_keywords
            else None
        )

        start = perf_counter()
        if self._event_manager:
            await self._event_manager.trigger(
                Event(
                    type=EventType.MODEL_MANAGER_CALL_BEFORE,
                    payload={
                        "engine_uri": engine_uri,
                        "modality": modality,
                        "operation": operation,
                    },
                    started=start,
                )
            )

        result: Any

        match modality:
            case Modality.AUDIO_CLASSIFICATION:
                assert (
                    operation.parameters["audio"]
                    and operation.parameters["audio"].path
                    and operation.parameters["audio"].sampling_rate
                )

                result = await model(
                    path=operation.parameters["audio"].path,
                    sampling_rate=operation.parameters["audio"].sampling_rate,
                )

            case Modality.AUDIO_SPEECH_RECOGNITION:
                assert (
                    operation.parameters["audio"]
                    and operation.parameters["audio"].path
                    and operation.parameters["audio"].sampling_rate
                )

                result = await model(
                    path=operation.parameters["audio"].path,
                    sampling_rate=operation.parameters["audio"].sampling_rate,
                )

            case Modality.AUDIO_TEXT_TO_SPEECH:
                assert (
                    operation.parameters["audio"]
                    and operation.parameters["audio"].path
                    and operation.parameters["audio"].sampling_rate
                )

                result = await model(
                    path=operation.parameters["audio"].path,
                    prompt=operation.input,
                    max_new_tokens=operation.generation_settings.max_new_tokens,
                    reference_path=operation.parameters[
                        "audio"
                    ].reference_path,
                    reference_text=operation.parameters[
                        "audio"
                    ].reference_text,
                    sampling_rate=operation.parameters["audio"].sampling_rate,
                )

            case Modality.AUDIO_GENERATION:
                assert (
                    operation.input
                    and operation.parameters["audio"]
                    and operation.parameters["audio"].path
                )

                result = await model(
                    operation.input,
                    operation.parameters["audio"].path,
                    operation.generation_settings.max_new_tokens,
                )

            case Modality.TEXT_GENERATION:
                assert operation.input and operation.parameters["text"]

                is_mlx = isinstance(model, MlxLmModel)
                if engine_uri.is_local and not is_mlx:
                    result = await model(
                        operation.input,
                        system_prompt=operation.parameters[
                            "text"
                        ].system_prompt,
                        settings=operation.generation_settings,
                        stopping_criterias=(
                            [stopping_criteria] if stopping_criteria else None
                        ),
                        manual_sampling=operation.parameters[
                            "text"
                        ].manual_sampling,
                        pick=operation.parameters["text"].pick_tokens,
                        skip_special_tokens=operation.parameters[
                            "text"
                        ].skip_special_tokens,
                        tool=tool,
                    )
                else:
                    result = await model(
                        operation.input,
                        system_prompt=operation.parameters[
                            "text"
                        ].system_prompt,
                        settings=operation.generation_settings,
                        tool=tool,
                    )

            case Modality.TEXT_QUESTION_ANSWERING:
                assert (
                    operation.input
                    and operation.parameters["text"]
                    and operation.parameters["text"].context
                )

                result = await model(
                    operation.input,
                    context=operation.parameters["text"].context,
                    system_prompt=operation.parameters["text"].system_prompt,
                )

            case Modality.TEXT_SEQUENCE_CLASSIFICATION:
                assert operation.input

                result = await model(operation.input)

            case Modality.TEXT_SEQUENCE_TO_SEQUENCE:
                assert operation.input and operation.parameters["text"]

                result = await model(
                    operation.input,
                    settings=operation.generation_settings,
                    stopping_criterias=(
                        [stopping_criteria] if stopping_criteria else None
                    ),
                )

            case Modality.TEXT_TOKEN_CLASSIFICATION:
                assert operation.input and operation.parameters["text"]

                result = await model(
                    operation.input,
                    labeled_only=operation.parameters["text"].labeled_only
                    or False,
                    system_prompt=operation.parameters["text"].system_prompt,
                )

            case Modality.TEXT_TRANSLATION:
                assert (
                    operation.input
                    and operation.parameters["text"]
                    and operation.parameters["text"].language_source
                    and operation.parameters["text"].language_destination
                )

                result = await model(
                    operation.input,
                    source_language=operation.parameters[
                        "text"
                    ].language_source,
                    destination_language=operation.parameters[
                        "text"
                    ].language_destination,
                    settings=operation.generation_settings,
                    stopping_criterias=(
                        [stopping_criteria] if stopping_criteria else None
                    ),
                    skip_special_tokens=operation.parameters[
                        "text"
                    ].skip_special_tokens,
                )

            case Modality.VISION_ENCODER_DECODER:
                assert (
                    operation.parameters["vision"]
                    and operation.parameters["vision"].path
                )

                result = await model(
                    operation.parameters["vision"].path,
                    prompt=operation.input,
                    skip_special_tokens=operation.parameters[
                        "vision"
                    ].skip_special_tokens,
                )

            case Modality.VISION_IMAGE_CLASSIFICATION:
                assert (
                    operation.parameters["vision"]
                    and operation.parameters["vision"].path
                )

                result = await model(operation.parameters["vision"].path)

            case Modality.VISION_IMAGE_TO_TEXT:
                assert (
                    operation.parameters["vision"]
                    and operation.parameters["vision"].path
                )

                result = await model(
                    operation.parameters["vision"].path,
                    skip_special_tokens=operation.parameters[
                        "vision"
                    ].skip_special_tokens,
                )

            case Modality.VISION_IMAGE_TEXT_TO_TEXT:
                assert (
                    operation.parameters["vision"]
                    and operation.parameters["vision"].path
                )

                result = await model(
                    operation.parameters["vision"].path,
                    operation.input,
                    system_prompt=operation.parameters["vision"].system_prompt,
                    settings=operation.generation_settings,
                    width=operation.parameters["vision"].width,
                )

            case Modality.VISION_OBJECT_DETECTION:
                assert (
                    operation.parameters["vision"]
                    and operation.parameters["vision"].path
                    and operation.parameters["vision"].threshold is not None
                )

                result = await model(
                    operation.parameters["vision"].path,
                    threshold=operation.parameters["vision"].threshold,
                )

            case Modality.VISION_TEXT_TO_IMAGE:
                assert (
                    operation.input
                    and operation.parameters["vision"]
                    and operation.parameters["vision"].path
                    and operation.parameters["vision"].color_model
                    and operation.parameters["vision"].high_noise_frac
                    is not None
                    and operation.parameters["vision"].image_format
                    and operation.parameters["vision"].n_steps is not None
                )

                result = await model(
                    operation.input,
                    operation.parameters["vision"].path,
                    color_model=operation.parameters["vision"].color_model,
                    high_noise_frac=operation.parameters[
                        "vision"
                    ].high_noise_frac,
                    image_format=operation.parameters["vision"].image_format,
                    n_steps=operation.parameters["vision"].n_steps,
                )

            case Modality.VISION_TEXT_TO_ANIMATION:
                assert (
                    operation.input
                    and operation.parameters["vision"]
                    and operation.parameters["vision"].path
                    and operation.parameters["vision"].n_steps is not None
                    and operation.parameters["vision"].timestep_spacing
                    and operation.parameters["vision"].beta_schedule
                    and operation.parameters["vision"].guidance_scale
                    is not None
                )

                result = await model(
                    operation.input,
                    operation.parameters["vision"].path,
                    beta_schedule=operation.parameters["vision"].beta_schedule,
                    guidance_scale=operation.parameters[
                        "vision"
                    ].guidance_scale,
                    steps=operation.parameters["vision"].n_steps,
                    timestep_spacing=operation.parameters[
                        "vision"
                    ].timestep_spacing,
                )

            case Modality.VISION_TEXT_TO_VIDEO:
                assert (
                    operation.input
                    and operation.parameters["vision"]
                    and operation.parameters["vision"].path
                )

                vision = operation.parameters["vision"]
                kwargs = {
                    "reference_path": vision.reference_path,
                    "negative_prompt": vision.negative_prompt,
                    "height": vision.height,
                    "downscale": vision.downscale,
                    "frames": vision.frames,
                    "denoise_strength": vision.denoise_strength,
                    "inference_steps": vision.inference_steps,
                    "decode_timestep": vision.decode_timestep,
                    "noise_scale": vision.noise_scale,
                    "frames_per_second": vision.frames_per_second,
                }
                if vision.width is not None:
                    kwargs["width"] = vision.width
                if vision.n_steps is not None:
                    kwargs["steps"] = vision.n_steps

                result = await model(
                    operation.input,
                    vision.path,
                    **kwargs,
                )

            case Modality.VISION_SEMANTIC_SEGMENTATION:
                assert (
                    operation.parameters["vision"]
                    and operation.parameters["vision"].path
                )

                result = await model(operation.parameters["vision"].path)

            case _:
                raise NotImplementedError(f"Modality {modality} not supported")

        end = perf_counter()
        if self._event_manager:
            await self._event_manager.trigger(
                Event(
                    type=EventType.MODEL_MANAGER_CALL_AFTER,
                    payload={
                        "engine_uri": engine_uri,
                        "modality": modality,
                        "operation": operation,
                        "result": result,
                    },
                    started=start,
                    finished=end,
                    elapsed=end - start,
                )
            )

        return result

    @staticmethod
    def get_operation_from_arguments(
        modality: Modality,
        args: Namespace,
        input_string: Input | None,
    ) -> Operation:
        reasoning_settings = ReasoningSettings(
            max_new_tokens=getattr(args, "reasoning_max_new_tokens", None),
            enabled=not getattr(args, "no_reasoning", False),
            stop_on_max_new_tokens=getattr(
                args,
                "reasoning_stop_on_max_new_tokens",
                False,
            ),
        )
        settings = GenerationSettings(
            do_sample=args.do_sample,
            enable_gradient_calculation=args.enable_gradient_calculation,
            max_new_tokens=args.max_new_tokens,
            max_length=getattr(args, "text_max_length", None),
            min_p=args.min_p,
            num_beams=getattr(args, "text_num_beams", None),
            repetition_penalty=args.repetition_penalty,
            temperature=args.temperature,
            top_k=args.top_k,
            top_p=args.top_p,
            use_cache=args.use_cache,
            chat_settings=ChatSettings(
                enable_thinking=not getattr(
                    args,
                    "chat_disable_thinking",
                    not reasoning_settings.enabled,
                )
            ),
            reasoning=reasoning_settings,
        )
        system_prompt = args.system or None

        requires_input = modality in {
            Modality.AUDIO_TEXT_TO_SPEECH,
            Modality.AUDIO_GENERATION,
            Modality.TEXT_GENERATION,
            Modality.TEXT_QUESTION_ANSWERING,
            Modality.TEXT_SEQUENCE_CLASSIFICATION,
            Modality.TEXT_SEQUENCE_TO_SEQUENCE,
            Modality.TEXT_TRANSLATION,
            Modality.TEXT_TOKEN_CLASSIFICATION,
            Modality.VISION_IMAGE_TEXT_TO_TEXT,
            Modality.VISION_TEXT_TO_IMAGE,
            Modality.VISION_TEXT_TO_ANIMATION,
            Modality.VISION_TEXT_TO_VIDEO,
        }

        match modality:
            case Modality.AUDIO_CLASSIFICATION:
                parameters = OperationParameters(
                    audio=OperationAudioParameters(
                        path=args.path,
                        sampling_rate=args.audio_sampling_rate,
                    )
                )

            case Modality.AUDIO_SPEECH_RECOGNITION:
                parameters = OperationParameters(
                    audio=OperationAudioParameters(
                        path=args.path,
                        sampling_rate=args.audio_sampling_rate,
                    )
                )

            case Modality.AUDIO_TEXT_TO_SPEECH:
                parameters = OperationParameters(
                    audio=OperationAudioParameters(
                        path=args.path,
                        reference_path=args.audio_reference_path,
                        reference_text=args.audio_reference_text,
                        sampling_rate=args.audio_sampling_rate,
                    )
                )

            case Modality.AUDIO_GENERATION:
                parameters = OperationParameters(
                    audio=OperationAudioParameters(
                        path=args.path,
                        sampling_rate=args.audio_sampling_rate,
                    )
                )

            case Modality.TEXT_QUESTION_ANSWERING:
                parameters = OperationParameters(
                    text=OperationTextParameters(
                        context=args.text_context,
                        system_prompt=system_prompt,
                    )
                )

            case Modality.TEXT_SEQUENCE_CLASSIFICATION:
                parameters = None

            case Modality.TEXT_SEQUENCE_TO_SEQUENCE:
                parameters = OperationParameters(
                    text=OperationTextParameters(
                        stop_on_keywords=args.stop_on_keyword,
                    )
                )

            case Modality.TEXT_TRANSLATION:
                parameters = OperationParameters(
                    text=OperationTextParameters(
                        language_destination=args.text_to_lang,
                        language_source=args.text_from_lang,
                        stop_on_keywords=args.stop_on_keyword,
                        skip_special_tokens=args.skip_special_tokens,
                    )
                )

            case Modality.TEXT_TOKEN_CLASSIFICATION:
                parameters = OperationParameters(
                    text=OperationTextParameters(
                        labeled_only=getattr(args, "text_labeled_only", None),
                        system_prompt=system_prompt,
                    )
                )

            case Modality.TEXT_GENERATION:
                parameters = OperationParameters(
                    text=OperationTextParameters(
                        manual_sampling=args.display_tokens or 0,
                        pick_tokens=(
                            10
                            if args.display_tokens and args.display_tokens > 0
                            else 0
                        ),
                        stop_on_keywords=args.stop_on_keyword,
                        skip_special_tokens=args.quiet
                        or args.skip_special_tokens,
                        system_prompt=system_prompt,
                    )
                )

            case Modality.VISION_IMAGE_CLASSIFICATION:
                parameters = OperationParameters(
                    vision=OperationVisionParameters(
                        path=args.path,
                    )
                )

            case (
                Modality.VISION_IMAGE_TO_TEXT | Modality.VISION_ENCODER_DECODER
            ):
                parameters = OperationParameters(
                    vision=OperationVisionParameters(
                        path=args.path,
                        skip_special_tokens=args.skip_special_tokens,
                    )
                )

            case Modality.VISION_IMAGE_TEXT_TO_TEXT:
                parameters = OperationParameters(
                    vision=OperationVisionParameters(
                        path=args.path,
                        system_prompt=system_prompt,
                        width=getattr(
                            args,
                            "vision_width",
                            getattr(args, "image_width", None),
                        ),
                    )
                )

            case Modality.VISION_OBJECT_DETECTION:
                parameters = OperationParameters(
                    vision=OperationVisionParameters(
                        path=args.path,
                        threshold=getattr(
                            args,
                            "vision_threshold",
                            getattr(args, "image_threshold", None),
                        ),
                    )
                )

            case Modality.VISION_TEXT_TO_IMAGE:
                parameters = OperationParameters(
                    vision=OperationVisionParameters(
                        path=args.path,
                        color_model=args.vision_color_model,
                        high_noise_frac=args.vision_high_noise_frac,
                        image_format=args.vision_image_format,
                        n_steps=args.vision_steps,
                    )
                )

            case Modality.VISION_TEXT_TO_ANIMATION:
                parameters = OperationParameters(
                    vision=OperationVisionParameters(
                        path=args.path,
                        n_steps=args.vision_steps,
                        timestep_spacing=args.vision_timestep_spacing,
                        beta_schedule=args.vision_beta_schedule,
                        guidance_scale=args.vision_guidance_scale,
                    )
                )

            case Modality.VISION_TEXT_TO_VIDEO:
                parameters = OperationParameters(
                    vision=OperationVisionParameters(
                        path=args.path,
                        reference_path=getattr(
                            args, "vision_reference_path", None
                        ),
                        negative_prompt=getattr(
                            args, "vision_negative_prompt", None
                        ),
                        width=getattr(args, "vision_width", None),
                        height=getattr(args, "vision_height", None),
                        downscale=getattr(args, "vision_downscale", None),
                        frames=getattr(args, "vision_frames", None),
                        denoise_strength=getattr(
                            args, "vision_denoise_strength", None
                        ),
                        n_steps=getattr(args, "vision_steps", None),
                        inference_steps=getattr(
                            args, "vision_inference_steps", None
                        ),
                        decode_timestep=getattr(
                            args, "vision_decode_timestep", None
                        ),
                        noise_scale=getattr(args, "vision_noise_scale", None),
                        frames_per_second=getattr(args, "vision_fps", None),
                    )
                )

            case Modality.VISION_SEMANTIC_SEGMENTATION:
                parameters = OperationParameters(
                    vision=OperationVisionParameters(
                        path=args.path,
                    )
                )

            case _:
                parameters = None

        operation = Operation(
            generation_settings=settings,
            input=input_string,
            modality=modality,
            parameters=parameters,
            requires_input=requires_input,
        )

        return operation

    def get_engine_settings(
        self,
        engine_uri: EngineUri,
        settings: dict | None = None,
        modality: Modality | None = None,
    ) -> TransformerEngineSettings:
        engine_settings_args = settings or {}

        if modality != Modality.EMBEDDING and not engine_uri.is_local:
            token = None
            if engine_uri.password and engine_uri.user:
                if engine_uri.user == "secret":
                    token = self._secrets.read(engine_uri.password)
                else:
                    token = None
            elif engine_uri.user:
                token = engine_uri.user

            if token:
                engine_settings_args.update(access_token=token)

        engine_settings = TransformerEngineSettings(**engine_settings_args)
        return engine_settings

    def load(
        self,
        engine_uri: EngineUri,
        modality: Modality = Modality.TEXT_GENERATION,
        *args,
        attention: AttentionImplementation | None = None,
        base_url: str | None = None,
        device: str | None = None,
        disable_loading_progress_bar: bool = False,
        loader_class: TextGenerationLoaderClass | None = "auto",
        backend: Backend = Backend.TRANSFORMERS,
        low_cpu_mem_usage: bool = False,
        parallel: ParallelStrategy | None = None,
        quiet: bool = False,
        output_hidden_states: bool | None = None,
        base_model_id: str | None = None,
        checkpoint: str | None = None,
        refiner_model_id: str | None = None,
        upsampler_model_id: str | None = None,
        revision: str | None = None,
        special_tokens: list[str] | None = None,
        subfolder: str | None = None,
        tokenizer: str | None = None,
        tokenizer_subfolder: str | None = None,
        tokens: list[str] | None = None,
        trust_remote_code: bool | None = None,
        weight_type: WeightType = "auto",
    ) -> ModelType:
        engine_settings_args = dict(
            base_url=base_url,
            cache_dir=self._hub.cache_dir,
            device=device,
            disable_loading_progress_bar=quiet or disable_loading_progress_bar,
            low_cpu_mem_usage=low_cpu_mem_usage,
            loader_class=loader_class,
            backend=backend,
            parallel=parallel,
            base_model_id=base_model_id or None,
            checkpoint=checkpoint or None,
            refiner_model_id=refiner_model_id or None,
            upsampler_model_id=upsampler_model_id or None,
            revision=revision,
            special_tokens=special_tokens or None,
            subfolder=subfolder or None,
            tokenizer_name_or_path=tokenizer,
            tokenizer_subfolder=tokenizer_subfolder or None,
            tokens=tokens or None,
            weight_type=weight_type,
        )

        if output_hidden_states is not None:
            engine_settings_args["output_hidden_states"] = output_hidden_states

        if modality != Modality.EMBEDDING:
            engine_settings_args.update(
                attention=attention or None,
                trust_remote_code=trust_remote_code or None,
            )

        engine_settings = self.get_engine_settings(
            engine_uri,
            engine_settings_args,
            modality=modality,
        )
        return self.load_engine(engine_uri, engine_settings, modality)

    def load_engine(
        self,
        engine_uri: EngineUri,
        engine_settings: TransformerEngineSettings,
        modality: Modality = Modality.TEXT_GENERATION,
    ) -> ModelType:
        assert isinstance(engine_uri, EngineUri)
        model_load_args = dict(
            model_id=engine_uri.model_id,
            settings=engine_settings,
            logger=self._logger,
        )

        # Load local model, or lazy-import per vendor
        if engine_uri.is_local:
            match modality:
                case Modality.EMBEDDING:
                    model = SentenceTransformerModel(**model_load_args)
                case Modality.AUDIO_CLASSIFICATION:
                    model = AudioClassificationModel(**model_load_args)
                case Modality.AUDIO_SPEECH_RECOGNITION:
                    model = SpeechRecognitionModel(**model_load_args)
                case Modality.AUDIO_TEXT_TO_SPEECH:
                    model = TextToSpeechModel(**model_load_args)
                case Modality.AUDIO_GENERATION:
                    model = AudioGenerationModel(**model_load_args)
                case Modality.VISION_OBJECT_DETECTION:
                    model = ObjectDetectionModel(**model_load_args)
                case Modality.VISION_IMAGE_CLASSIFICATION:
                    model = ImageClassificationModel(**model_load_args)
                case Modality.VISION_IMAGE_TO_TEXT:
                    model = ImageToTextModel(**model_load_args)
                case Modality.VISION_IMAGE_TEXT_TO_TEXT:
                    model = ImageTextToTextModel(**model_load_args)
                case Modality.VISION_ENCODER_DECODER:
                    model = VisionEncoderDecoderModel(**model_load_args)
                case Modality.VISION_TEXT_TO_IMAGE:
                    model = TextToImageModel(**model_load_args)
                case Modality.VISION_TEXT_TO_ANIMATION:
                    model = TextToAnimationModel(**model_load_args)
                case Modality.VISION_TEXT_TO_VIDEO:
                    model = TextToVideoModel(**model_load_args)
                case Modality.VISION_SEMANTIC_SEGMENTATION:
                    model = SemanticSegmentationModel(**model_load_args)
                case Modality.TEXT_QUESTION_ANSWERING:
                    model = QuestionAnsweringModel(**model_load_args)
                case Modality.TEXT_SEQUENCE_CLASSIFICATION:
                    model = SequenceClassificationModel(**model_load_args)
                case Modality.TEXT_SEQUENCE_TO_SEQUENCE:
                    model = SequenceToSequenceModel(**model_load_args)
                case Modality.TEXT_TRANSLATION:
                    model = TranslationModel(**model_load_args)
                case Modality.TEXT_TOKEN_CLASSIFICATION:
                    model = TokenClassificationModel(**model_load_args)
                case _:
                    match engine_settings.backend:
                        case Backend.MLXLM:
                            from ..model.nlp.text.mlxlm import MlxLmModel

                            model = MlxLmModel(**model_load_args)
                        case Backend.VLLM:
                            from ..model.nlp.text.vllm import VllmModel

                            model = VllmModel(**model_load_args)
                        case _:
                            model = TextGenerationModel(**model_load_args)
        elif (
            modality == Modality.TEXT_GENERATION
            and engine_uri.vendor == "openai"
        ):
            from ..model.nlp.text.vendor.openai import OpenAIModel

            model = OpenAIModel(**model_load_args)
        elif (
            modality == Modality.TEXT_GENERATION
            and engine_uri.vendor == "openrouter"
        ):
            from ..model.nlp.text.vendor.openrouter import OpenRouterModel

            model = OpenRouterModel(**model_load_args)
        elif (
            modality == Modality.TEXT_GENERATION
            and engine_uri.vendor == "anyscale"
        ):
            from ..model.nlp.text.vendor.anyscale import AnyScaleModel

            model = AnyScaleModel(**model_load_args)
        elif (
            modality == Modality.TEXT_GENERATION
            and engine_uri.vendor == "together"
        ):
            from ..model.nlp.text.vendor.together import TogetherModel

            model = TogetherModel(**model_load_args)
        elif (
            modality == Modality.TEXT_GENERATION
            and engine_uri.vendor == "deepseek"
        ):
            from ..model.nlp.text.vendor.deepseek import DeepSeekModel

            model = DeepSeekModel(**model_load_args)
        elif (
            modality == Modality.TEXT_GENERATION
            and engine_uri.vendor == "deepinfra"
        ):
            from ..model.nlp.text.vendor.deepinfra import DeepInfraModel

            model = DeepInfraModel(**model_load_args)
        elif (
            modality == Modality.TEXT_GENERATION
            and engine_uri.vendor == "groq"
        ):
            from ..model.nlp.text.vendor.groq import GroqModel

            model = GroqModel(**model_load_args)
        elif (
            modality == Modality.TEXT_GENERATION
            and engine_uri.vendor == "ollama"
        ):
            from ..model.nlp.text.vendor.ollama import OllamaModel

            model = OllamaModel(**model_load_args)
        elif (
            modality == Modality.TEXT_GENERATION
            and engine_uri.vendor == "huggingface"
        ):
            from ..model.nlp.text.vendor.huggingface import HuggingfaceModel

            model = HuggingfaceModel(**model_load_args)
        elif (
            modality == Modality.TEXT_GENERATION
            and engine_uri.vendor == "hyperbolic"
        ):
            from ..model.nlp.text.vendor.hyperbolic import HyperbolicModel

            model = HyperbolicModel(**model_load_args)
        elif (
            modality == Modality.TEXT_GENERATION
            and engine_uri.vendor == "litellm"
        ):
            from ..model.nlp.text.vendor.litellm import LiteLLMModel

            model = LiteLLMModel(**model_load_args)
        else:
            raise NotImplementedError()

        self._stack.enter_context(model)
        return model

    @staticmethod
    def parse_uri(uri: str) -> EngineUri:
        parsed = urlparse(uri)
        if not parsed.scheme:
            uri = f"ai://{uri}"
            parsed = urlparse(uri)

        if parsed.scheme != "ai":
            raise ValueError(
                f"Invalid scheme {parsed.scheme!r}, expected 'ai'"
            )

        vendor = parsed.hostname
        if not vendor or vendor not in get_args(Vendor) or vendor == "local":
            vendor = None
        use_host = bool(vendor)
        path_prefixed = parsed.path.startswith("/")
        params = dict(parse_qsl(parsed.query))

        # urlparse() normalizes hostname to lowercase, so keep original case
        authority = parsed.netloc.rsplit("@", 1)[-1]
        hostname = authority.split(":", 1)[0]

        model_id = (
            hostname + ("/" if path_prefixed else "")
            if not vendor and hostname != "local"
            else ""
        ) + (parsed.path[1:] if path_prefixed else parsed.path)
        engine_uri = EngineUri(
            vendor=vendor,
            host=hostname if use_host else None,
            port=(parsed.port or None) if use_host else None,
            user=parsed.username or None,
            password=parsed.password or None,
            model_id=model_id,
            params=params,
        )
        return engine_uri



"""Parser emitting events for detected tool calls."""

from io import StringIO
from time import perf_counter
from typing import Any, Iterable

from ....entities import ToolCallToken
from ....event import Event, EventType
from ....event.manager import EventManager
from ....tool.manager import ToolManager


class ToolCallParser:
    """Parse tool calls during streaming."""

    def __init__(
        self, tool_manager: ToolManager, event_manager: EventManager | None
    ) -> None:
        self._tool_manager = tool_manager
        self._event_manager = event_manager
        self._buffer = StringIO()
        self._tag_buffer = ""
        self._inside_call = False

    async def push(self, token_str: str) -> Iterable[Any]:
        buffer_value = self._buffer.getvalue()
        should_check = self._tool_manager.is_potential_tool_call(
            buffer_value, token_str
        )

        prev_inside = self._inside_call

        self._buffer.write(token_str)
        self._tag_buffer += token_str
        if len(self._tag_buffer) > 64:
            self._tag_buffer = self._tag_buffer[-64:]

        if not self._inside_call and (
            "<tool_call" in self._tag_buffer
            or "<tool " in self._tag_buffer
            or "<tool>" in self._tag_buffer
        ):
            self._inside_call = True

        if self._inside_call and (
            "</tool_call>" in self._tag_buffer
            or "</tool>" in self._tag_buffer
            or "/>" in self._tag_buffer
        ):
            self._inside_call = False

        start_triggered = not prev_inside and self._inside_call

        item = (
            ToolCallToken(token_str)
            if prev_inside or start_triggered
            else token_str
        )

        if not should_check:
            return [item]

        if self._event_manager:
            await self._event_manager.trigger(
                Event(type=EventType.TOOL_DETECT)
            )

        calls = self._tool_manager.get_calls(self._buffer.getvalue())
        if not calls:
            return [item]

        event = Event(
            type=EventType.TOOL_PROCESS, payload=calls, started=perf_counter()
        )

        self._buffer = StringIO()
        self._tag_buffer = ""
        self._inside_call = False
        return [item, event]

    async def flush(self) -> Iterable[Any]:
        return []



from argparse import Namespace
from ...cli import get_input
from ...entities import Token, TransformerEngineSettings
from ...model.hubs.huggingface import HuggingfaceHub
from ...model.nlp.text.generation import TextGenerationModel
from logging import Logger
from rich.console import Console
from rich.theme import Theme


async def tokenize(
    args: Namespace,
    console: Console,
    theme: Theme,
    hub: HuggingfaceHub,
    logger: Logger,
) -> list[Token] | None:
    assert args.tokenizer

    _, _i, _n = theme._, theme.icons, theme._n

    tokenizer_name_or_path = args.tokenizer
    with TextGenerationModel(
        tokenizer_name_or_path,
        settings=TransformerEngineSettings(
            device=args.device,
            cache_dir=hub.cache_dir,
            subfolder=getattr(args, "subfolder", None),
            tokenizer_subfolder=getattr(args, "tokenizer_subfolder", None),
            tokenizer_name_or_path=tokenizer_name_or_path,
            tokens=(
                args.token
                if args.token and isinstance(args.token, list)
                else None
            ),
            special_tokens=(
                args.special_token
                if args.special_token and isinstance(args.special_token, list)
                else None
            ),
            auto_load_model=False,
            auto_load_tokenizer=True,
            disable_loading_progress_bar=args.disable_loading_progress_bar,
            low_cpu_mem_usage=args.low_cpu_mem_usage,
            loader_class=args.loader_class,
            backend=args.backend,
            weight_type=args.weight_type,
        ),
        logger=logger,
    ) as lm:
        logger.debug("Loaded tokenizer %s", lm.tokenizer_config.__repr__())
        console.print(theme.tokenizer_config(lm.tokenizer_config))

        if args.save:
            paths = lm.save_tokenizer(args.save)
            total_files = len(paths)
            console.print(theme.saved_tokenizer_files(args.save, total_files))
            return

        input_string = get_input(
            console,
            _i["user_input"] + " ",
            echo_stdin=not args.no_repl,
            is_quiet=args.quiet,
        )
        if input_string:
            logger.debug("Loaded model %s", lm.config.__repr__())
            tokens = lm.tokenize(input_string)

            panel = theme.tokenizer_tokens(
                tokens,
                lm.tokenizer_config.tokens,
                lm.tokenizer_config.special_tokens,
                display_details=True,
            )
            console.print(panel)



from . import Tool, ToolSet
from .parser import ToolCallParser
from ..entities import (
    ToolCall,
    ToolCallContext,
    ToolCallResult,
    ToolManagerSettings,
    ToolFilter,
    ToolTransformer,
)
from collections.abc import Callable, Sequence
from contextlib import AsyncExitStack, ContextDecorator
from uuid import uuid4


class ToolManager(ContextDecorator):
    _parser: ToolCallParser
    _stack: AsyncExitStack
    _tools: dict[str, Callable] | None = None
    _toolsets: Sequence[ToolSet] | None = None

    @staticmethod
    def _matches_namespace(tool_name: str, namespace: str | None) -> bool:
        if not namespace:
            return True
        return tool_name == namespace or tool_name.startswith(f"{namespace}.")

    @classmethod
    def create_instance(
        cls,
        *,
        available_toolsets: Sequence[ToolSet] | None = None,
        enable_tools: list[str] | None = None,
        settings: ToolManagerSettings | None = None,
    ):
        parser = ToolCallParser(
            eos_token=settings.eos_token if settings else None,
            tool_format=settings.tool_format if settings else None,
        )
        return cls(
            available_toolsets=available_toolsets,
            enable_tools=enable_tools,
            parser=parser,
            settings=settings,
        )

    @property
    def is_empty(self) -> bool:
        return not bool(self._tools)

    @property
    def tools(self) -> list[Callable] | None:
        return list(self._tools.values()) if self._tools else None

    def json_schemas(self) -> list[dict] | None:
        schemas = []
        for toolset in self._toolsets:
            schemas.extend(toolset.json_schemas())
        return schemas

    def __init__(
        self,
        *,
        available_toolsets: Sequence[ToolSet] | None = None,
        enable_tools: list[str] | None = None,
        parser: ToolCallParser,
        settings: ToolManagerSettings | None = None,
    ):
        self._parser = parser
        self._settings = settings or ToolManagerSettings()
        self._stack = AsyncExitStack()

        enabled_toolsets = []
        if available_toolsets:
            for toolset in available_toolsets:
                if enable_tools is not None:
                    toolset = toolset.with_enabled_tools(enable_tools)
                if toolset.tools:
                    enabled_toolsets.append(toolset)

        self._tools = {}
        if enabled_toolsets:
            for i, toolset in enumerate(enabled_toolsets):
                prefix = f"{toolset.namespace}." if toolset.namespace else ""
                for tool in toolset.tools:
                    name = getattr(tool, "__name__", tool.__class__.__name__)
                    self._tools[f"{prefix}{name}"] = tool

        self._toolsets = enabled_toolsets

    def set_eos_token(self, eos_token: str) -> None:
        self._parser.set_eos_token(eos_token)

    def is_potential_tool_call(self, buffer: str, token_str: str) -> bool:
        """Proxy :meth:`ToolCallParser.is_potential_tool_call`."""
        return self._parser.is_potential_tool_call(buffer, token_str)

    def get_calls(self, text: str) -> list[ToolCall] | None:
        return self._parser(text)

    async def __aenter__(self) -> "ToolManager":
        if self._toolsets:
            for i, toolset in enumerate(self._toolsets):
                toolset = await self._stack.enter_async_context(toolset)
                self._toolsets[i] = toolset
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: BaseException | None,
    ) -> bool:
        return await self._stack.__aexit__(exc_type, exc_value, traceback)

    async def __call__(
        self, call: ToolCall, context: ToolCallContext
    ) -> ToolCallResult | None:
        """Execute a single tool call and return the result."""
        assert call

        history = context.calls or []

        if self._settings.avoid_repetition and history:
            last = history[-1]
            if last.name == call.name and last.arguments == call.arguments:
                return None

        if (
            self._settings.maximum_depth is not None
            and len(history) + 1 > self._settings.maximum_depth
        ):
            return None

        tool = self._tools.get(call.name, None) if self._tools else None

        if not tool:
            return None

        if self._settings.filters:
            for f in self._settings.filters:
                namespace = None
                func = f
                if isinstance(f, ToolFilter):
                    func = f.func
                    namespace = f.namespace
                if not self._matches_namespace(call.name, namespace):
                    continue
                modified = func(call, context)
                if modified is not None:
                    assert isinstance(modified, tuple) and len(modified) == 2
                    call, context = modified

        is_native_tool = isinstance(tool, Tool)

        result = (
            await tool(*call.arguments.values(), context=context)
            if is_native_tool and call.arguments
            else (
                tool(context=context)
                if is_native_tool
                else (
                    await tool(*call.arguments.values())
                    if call.arguments
                    else tool()
                )
            )
        )

        if self._settings.transformers:
            for t in self._settings.transformers:
                namespace = None
                func = t
                if isinstance(t, ToolTransformer):
                    func = t.func
                    namespace = t.namespace
                if not self._matches_namespace(call.name, namespace):
                    continue
                transformed = func(call, context, result)
                if transformed is not None:
                    result = transformed

        return ToolCallResult(
            id=uuid4(),
            call=call,
            name=call.name,
            arguments=call.arguments,
            result=result,
        )



from ...memory.partitioner import Encoding, PartitionerException
from ...utils import _j
from dataclasses import dataclass
from logging import Logger
from tree_sitter_python import language as python
from tree_sitter import Language, Node, Parser
from typing import Literal

LanguageName = Literal["python"]

SymbolType = Literal[
    "class",
    "function",
]

ParameterType = Literal[
    "default_parameter",
    "dictionary_splat_pattern",
    "identifier",
    "keyword_separator",
    "typed_default_parameter",
    "typed_parameter",
]


@dataclass(frozen=True, kw_only=True)
class Parameter:
    parameter_type: ParameterType
    name: str
    type: str | None


@dataclass(frozen=True, kw_only=True)
class Function:
    id: str
    namespace: str | None
    class_name: str | None
    name: str
    parameters: list[Parameter] | None
    return_type: str | None


@dataclass(frozen=True, kw_only=True)
class Symbol:
    symbol_type: SymbolType
    id: str


@dataclass(frozen=True, kw_only=True)
class CodePartition:
    data: str
    encoding: Encoding
    symbols: list[Symbol]


class CodePartitioner:
    _logger: Logger
    _parsers: dict[LanguageName, tuple[Parser, Language]] = {}

    def __init__(self, logger: Logger):
        self._logger = logger

    def partition(
        self,
        language_name: LanguageName,
        input: str,
        encoding: Encoding,
        max_chars: int,
        namespace: str | None = None,
    ) -> tuple[list[CodePartition], list[Function] | None]:
        parser, language = self._get_parser(language_name)
        tree = parser.parse(input.encode(encoding), encoding=encoding)
        root_node = tree.root_node

        self._logger.debug("Parsing %s code for functions", language_name)
        functions = self._get_functions(namespace, root_node, encoding)

        if root_node.children and root_node.children[0].is_error:
            error_node = root_node.children[0]
            error_name = error_node.grammar_name
            error_message = error_node.text.decode(encoding)
            error_row, error_column = error_node.start_point
            raise PartitionerException(
                f'{error_name}: "{error_message}" at'
                f" {error_row},{error_column}"
            )

        self._logger.debug("Partitioning %s code", language_name)
        partitions = self._partition(
            input, encoding, max_chars, root_node, current_namespace=namespace
        )
        self._logger.debug(
            "Partitioned %s into %d partitions and %d functions",
            language_name,
            len(partitions),
            len(functions),
        )
        return partitions, functions

    def _get_parser(
        self, language_name: LanguageName
    ) -> tuple[Parser, Language]:
        if language_name in self._parsers:
            return self._parsers[language_name]

        assert language_name == "python"  # @TODO obviously expand this
        language_impl = python()
        language = Language(language_impl)
        parser = Parser(language)
        self._parsers[language_name] = (parser, language)
        return parser, language

    # Implementation inspiration from Kevin Lu / SweepAI
    @classmethod
    def _partition(
        cls,
        source: str,
        encoding: Encoding,
        max_chars: int,
        node: Node,
        last_end: int = 0,
        current_namespace: str | None = None,
        current_class_name: str | None = None,
        current_symbols: list[Symbol] | None = None,
    ) -> list[CodePartition]:
        if current_symbols is None:
            current_symbols = []

        chunk, chunks = "", []

        for child_node in node.children:
            if child_node.type == "class_definition":
                class_name_node = child_node.child_by_field_name("name")
                current_class_name = (
                    class_name_node.text.decode(encoding)
                    if class_name_node
                    else None
                )
                class_id = _j(".", [current_namespace, current_class_name])
                child_symbols = current_symbols + [
                    Symbol(symbol_type="class", id=class_id)
                ]
            elif child_node.type in (
                "function_definition",
                "async_function_definition",
            ):
                function_id, _ = cls._get_function_id_and_name_from_node(
                    current_namespace, current_class_name, child_node, encoding
                )
                child_symbols = current_symbols + [
                    Symbol(symbol_type="function", id=function_id)
                ]
            else:
                child_symbols = current_symbols

            start, end = child_node.start_byte, child_node.end_byte
            length = end - start
            if length > max_chars:
                if chunk:
                    chunks.append(
                        CodePartition(
                            data=chunk,
                            encoding=encoding,
                            symbols=child_symbols or current_symbols,
                        )
                    )
                child_chunks = cls._partition(
                    source,
                    encoding,
                    max_chars,
                    child_node,
                    last_end,
                    current_namespace,
                    current_class_name,
                    child_symbols,
                )
                chunks.extend(child_chunks)
                chunk, last_end = "", end
                continue
            fragment = source[last_end:end]
            if len(chunk) + length > max_chars:
                if chunk:
                    chunks.append(
                        CodePartition(
                            data=chunk,
                            encoding=encoding,
                            symbols=child_symbols or current_symbols,
                        )
                    )
                chunk = fragment
            else:
                chunk += fragment
            last_end = end
        if chunk:
            chunks.append(
                CodePartition(
                    data=chunk,
                    encoding=encoding,
                    symbols=child_symbols or current_symbols,
                )
            )
        return chunks

    @classmethod
    def _get_functions(
        cls,
        current_namespace: str | None,
        node: Node,
        encoding: Encoding,
        current_class_name: str | None = None,
    ) -> list[Function]:
        assert node and encoding
        results = []
        if node.type == "class_definition":
            class_name_node = node.child_by_field_name("name")
            class_name = (
                class_name_node.text.decode(encoding)
                if class_name_node
                else None
            )
            for child in node.children:
                functs = cls._get_functions(
                    current_namespace,
                    child,
                    encoding,
                    class_name,
                )
                results.extend(functs)
        elif node.type in ("function_definition", "async_function_definition"):
            funct = cls._get_function_from_node(
                current_namespace, current_class_name, node, encoding
            )
            results.append(funct)
        else:
            for child in node.children:
                functs = cls._get_functions(
                    current_namespace,
                    child,
                    encoding,
                    current_class_name,
                )
                results.extend(functs)

        return results

    @classmethod
    def _get_function_from_node(
        cls,
        current_namespace: str | None,
        current_class_name: str | None,
        node: Node,
        encoding: Encoding,
    ) -> Function:
        assert node.type in (
            "function_definition",
            "async_function_definition",
        )
        function_id, function_name = cls._get_function_id_and_name_from_node(
            current_namespace, current_class_name, node, encoding
        )
        params_node = node.child_by_field_name("parameters")
        return_type_node = node.child_by_field_name("return_type")
        return Function(
            id=function_id,
            namespace=current_namespace,
            class_name=current_class_name,
            name=function_name,
            parameters=(
                cls._get_parameters(params_node, encoding)
                if params_node
                else None
            ),
            return_type=(
                return_type_node.text.decode(encoding)
                if return_type_node
                else None
            ),
        )

    @staticmethod
    def _get_function_id_and_name_from_node(
        current_namespace: str | None,
        current_class_name: str | None,
        node: Node,
        encoding: Encoding,
    ) -> tuple[str, str]:
        assert node.type in (
            "function_definition",
            "async_function_definition",
        )
        function_name_node = node.child_by_field_name("name")
        assert function_name_node
        function_name = function_name_node.text.decode(encoding)
        assert function_name
        function_id = _j(
            ".", [current_namespace, current_class_name, function_name]
        )
        return function_id, function_name

    @staticmethod
    def _get_parameters(
        node: Node, encoding: Encoding
    ) -> list[dict[str, str | None]]:
        assert node
        parameters = []
        for child in node.children:
            if child.type in (
                "default_parameter",
                "typed_default_parameter",
                "typed_parameter",
            ):
                param_type = child.child_by_field_name("type").text.decode(
                    encoding
                )
                param_name = None
                if child.named_child_count:
                    for parameter_child in child.children:
                        if parameter_child.type == "identifier":
                            param_name = parameter_child.text.decode(encoding)
                            break
                parameters.append(
                    Parameter(
                        parameter_type=child.type,
                        name=param_name,
                        type=param_type,
                    )
                )
            elif child.type in (
                "dictionary_splat_pattern",
                "identifier",
                "keyword_separator",
            ):
                parameters.append(
                    Parameter(
                        parameter_type=child.type,
                        name=child.text.decode(encoding),
                        type=None,
                    )
                )

        return parameters
